
/*******************************************************************************
*
* File plaq1.c
*
* Copyright (C) 2015, 2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Calculation of the quadratic part of the Wilson plaquette action.
*
*   void plaq1(sl3_alg_dble *ad,double **plaq)
*     Computes the leading-order contribution to the plaquette loop in
*     all 6 planes as a function of the time x0. The result is averaged
*     over the space coordinates and assigned to the plaq array.
*
* The calculated plaquette observable in the (mu,nu)-plane is given by
*
*   (1/4)*Re{<F^a_{mu,nu}(x)F^a_{mu,nu}(x)},
*
*   F_{mu,nu}(x)=A_nu(x+mu)-A_nu(x)-A_mu(x+nu)+A_mu(x),
*
* where A_mu(x) is the gauge potential defined by the field array ad.
*
* The program plaq1() assigns the calculated values to
*
*   plaq[n][x0], n=0,..,5, x0=0,..,T-1,
*
* where n counts the planes (mu,nu)=[(0,1),(0,2),(0,3),(1,2),(1,3),(2,3)] and
* T=NPROC0*L0.
*
* The program plaq1() acts globally and must be called on all MPI processes
* simultaneously.
*
*******************************************************************************/

#define PLAQ1_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "su3fcts.h"
#include "pmdb.h"
#include "lattice.h"
#include "aflds.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static double loc_plaq[6][N0];
static sl3_alg_dble Xw ALIGNED16;


static void local_plaq(sl3_alg_dble *ad)
{
   int n,ix,x0,ip[4];
   complex_dble w;

   for (n=0;n<6;n++)
   {
      for (x0=0;x0<N0;x0++)
         loc_plaq[n][x0]=0.0;
   }

   for (ix=0;ix<VOLUME;ix++)
   {
      x0=global_time(ix);

      for (n=0;n<6;n++)
      {
         plaq_aidx(n,ix,ip);

         Xw=ad[ip[1]];
         sl3_sub(ad+ip[2],&Xw);
         sl3_sub(ad+ip[3],&Xw);
         sl3_add(ad+ip[0],&Xw);

         w=sl3_biform(&Xw,&Xw);
         loc_plaq[n][x0]+=w.re;
      }
   }
}


void plaq1(sl3_alg_dble *ad,double **plaq)
{
   int n,x0;
   double r0;

   copy_bnd_ad(0,ad);
   local_plaq(ad);
   r0=1.0/(4.0*(double)(N1*N2*N3));

   for (n=0;n<6;n++)
   {
      for (x0=0;x0<N0;x0++)
         loc_plaq[n][x0]*=r0;

      MPI_Reduce(loc_plaq[n],plaq[n],N0,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast(plaq[n],N0,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }
}
