
/*******************************************************************************
*
* File AFPn_action.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Calculation of the homogeneous parts of the Faddeev-Popov action.
*
*   complex_dble AFPn_action(n,sl3_alg_dble *ad)
*     Returns the contribution of degree n in the gauge potential ad
*     to the measure action.
*
* The program AFPn_action() acts globally and must be called on all MPI
* processes simultaneously.
*
*******************************************************************************/

#define AFPN_ACTION_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "su3fcts.h"
#include "lattice.h"
#include "aflds.h"
#include "cflds.h"
#include "vxfcts.h"
#include "global.h"

#define N0 (NPROC0*L0)

static int idsm,ns=0;
static su3_dble *eX,*Xw,*Yw,*dCb,*dC,*C;


static void alloc_arrays(int n)
{
   if (ns==0)
      idsm=init_hsum(2);

   if (n>ns)
   {
      if (ns>0)
         afree(eX);

      eX=amalloc((n+6)*sizeof(*eX),ALIGN);
      error(eX==NULL,1,"alloc_arrays [AFP_action.c]",
            "Unable to allocate auxiliary arrays");
      ns=n;
   }

   Xw=eX+n+1;
   Yw=Xw+1;
   dCb=Yw+1;
   dC=dCb+1;
   C=dC+1;

   cm3x3_zero(n+6,eX);
}


static void exp_to_poly(int n,sl3_alg_dble *X)
{
   int i;
   double r;

   cm3x3_unity(1,eX);
   sl3_to_su3(X,Xw);

   for (i=1;i<=n;i++)
   {
      su3xsu3(Xw,eX+i-1,eX+i);
      r=1.0/(double)(i);
      cm3x3_mulr(&r,eX+i,eX+i);
   }
}


static void grad_cd(sl3_alg_dble *cd,sl3_alg_dble *ad)
{
   int ix,iy,t,mu;

   cpcd_int_bnd(0,cd);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      if (t==0)
      {
         iy=iup[ix][0];
         (*ad)=cd[iy];
         _sl3_alg_sub_assign(*ad,cd[ix]);
         ad+=1;

         iy=idn[ix][0];
         _sl3_alg_mulr_assign(*ad,-1.0,cd[iy]);
         ad+=1;
      }
      else if (t==(N0-1))
      {
         _sl3_alg_mulr_assign(*ad,-1.0,cd[ix]);
         ad+=1;

         iy=idn[ix][0];
         (*ad)=cd[ix];
         _sl3_alg_sub_assign(*ad,cd[iy]);
         ad+=1;
      }
      else
      {
         iy=iup[ix][0];
         (*ad)=cd[iy];
         _sl3_alg_sub_assign(*ad,cd[ix]);
         ad+=1;

         iy=idn[ix][0];
         (*ad)=cd[ix];
         _sl3_alg_sub_assign(*ad,cd[iy]);
         ad+=1;
      }

      for (mu=1;mu<4;mu++)
      {
         iy=iup[ix][mu];
         (*ad)=cd[iy];
         _sl3_alg_sub_assign(*ad,cd[ix]);
         ad+=1;

         iy=idn[ix][mu];
         (*ad)=cd[ix];
         _sl3_alg_sub_assign(*ad,cd[iy]);
         ad+=1;
      }
   }
}


static void local_action(int n,sl3_alg_dble *ad,
                         sl3_alg_dble *cbd,sl3_alg_dble *cd)
{
   int ix,t,ifc,k;
   double r;
   complex_dble z0;
   sl3_alg_dble **wad,*gcbd,*gcd;

   wad=reserve_wad(2);
   gcbd=wad[0];
   gcd=wad[1];
   grad_cd(cbd,gcbd);
   grad_cd(cd,gcd);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         exp_to_poly(n,ad);
         sl3_to_su3(gcbd,dCb);
         sl3_to_su3(gcd,dC);

         su3xsu3(dC,eX+n,Xw);

         for (k=1;k<=n;k++)
         {
            su3xsu3(dC,eX+n-k,Yw);

            if (k&0x1)
               cm3x3_mul_sub(eX+k,Yw,Xw);
            else
               cm3x3_mul_add(eX+k,Yw,Xw);
         }

         r=bernoulli(n);
         cm3x3_mulr(&r,Xw,Xw);

         if (n==1)
         {
            if (ifc&0x1)
               sl3_to_su3(cd+idn[ix][ifc/2],C);
            else
               sl3_to_su3(cd+ix,C);

            su3xsu3(eX+1,C,Yw);
            cm3x3_mul_sub(C,eX+1,Yw);
            cm3x3_add(Yw,Xw);
         }

         su3xsu3(dCb,Xw,Xw);

         z0.re=(*Xw).c11.re+(*Xw).c22.re+(*Xw).c33.re;
         z0.im=(*Xw).c11.im+(*Xw).c22.im+(*Xw).c33.im;

         if ((t==0)&&(ifc>1))
            r=-1.0;
         else
            r=-2.0;

         z0.re*=r;
         z0.im*=r;

         add_to_hsum(idsm,(double*)(&z0));

         ad+=1;
         gcbd+=1;
         gcd+=1;
      }
   }

   release_wad();
}


complex_dble AFPn_action(int n,sl3_alg_dble *ad,
                         sl3_alg_dble *cdb,sl3_alg_dble *cd)
{
   complex_dble z;

   z.re=0.0;
   z.im=0.0;

   if ((n<1)||((n>1)&&(n&0x1)))
      return z;

   alloc_arrays(n);
   reset_hsum(idsm);
   local_action(n,ad,cdb,cd);
   global_hsum(idsm,(double*)(&z));

   return z;
}
