
/*******************************************************************************
*
* File AGn_action.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Calculation of the homogeneous parts of the Wilson plaquette action.
*
*   complex_dble AGn_action(n,sl3_alg_dble *ad)
*     Returns the contribution of degree n in the gauge potential ad
*     to the Wilson plaquette action. The factor 1/g0^2 is omitted.
*
* The program AGn_action() acts globally and must be called on all MPI
* processes simultaneously.
*
*******************************************************************************/

#define AGN_ACTION_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "pmdb.h"
#include "lattice.h"
#include "su3fcts.h"
#include "aflds.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static int idsm,ns=0;
static su3_dble *M0,*M1,*Mw,*eA[4];
static su3_dble Xw ALIGNED16;


static void alloc_arrays(int n)
{
   if (ns>0)
      afree(M0);
   else
      idsm=init_hsum(2);

   M0=amalloc(7*(n+1)*sizeof(*M0),ALIGN);
   error(M0==NULL,1,"alloc_arrays [AGn_action.c]",
         "Unable to allocate auxiliary arrays");

   M1=M0+n+1;
   Mw=M1+n+1;
   eA[0]=Mw+n+1;
   eA[1]=eA[0]+n+1;
   eA[2]=eA[1]+n+1;
   eA[3]=eA[2]+n+1;

   cm3x3_zero(7*(n+1),M0);
   ns=n;
}


static void exp_to_poly(int n,sl3_alg_dble *X,su3_dble *eX)
{
   int i;
   double r;

   cm3x3_unity(1,eX);
   sl3_to_su3(X,&Xw);

   for (i=1;i<=n;i++)
   {
      su3xsu3(&Xw,eX+i-1,eX+i);
      r=1.0/(double)(i);
      cm3x3_mulr(&r,eX+i,eX+i);
   }
}


static void flip_sign(int n,su3_dble *eX)
{
   int i;
   double sgn;

   sgn=-1.0;

   for (i=1;i<=n;i++)
   {
      if (i&0x1)
         cm3x3_mulr(&sgn,eX+i,eX+i);
   }
}


static void local_action(int n,sl3_alg_dble *ad)
{
   int ix,t,np,ip[4];
   double wc;
   complex_dble z0,z1;

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (np=0;np<6;np++)
      {
         plaq_aidx(np,ix,ip);

         exp_to_poly(n,ad+ip[0],eA[0]);
         exp_to_poly(n,ad+ip[1],eA[1]);
         exp_to_poly(n,ad+ip[2],eA[2]);
         exp_to_poly(n,ad+ip[3],eA[3]);
         flip_sign(n,eA[2]);
         flip_sign(n,eA[3]);

         cm3x3_mul_poly(n,eA[3],eA[2],M0);
         cm3x3_mul_poly(n,eA[1],M0,Mw);
         cm3x3_mul_poly(n,eA[0],Mw,M0);

         cm3x3_mul_poly(n,eA[1],eA[0],M1);
         cm3x3_mul_poly(n,eA[3],M1,Mw);
         cm3x3_mul_poly(n,eA[2],Mw,M1);

         z0.re=M0[n].c11.re+M0[n].c22.re+M0[n].c33.re;
         z0.im=M0[n].c11.im+M0[n].c22.im+M0[n].c33.im;

         z1.re=M1[n].c11.re+M1[n].c22.re+M1[n].c33.re;
         z1.im=M1[n].c11.im+M1[n].c22.im+M1[n].c33.im;

         if (n&0x1)
         {
            z0.re-=z1.re;
            z0.im-=z1.im;
         }
         else
         {
            z0.re+=z1.re;
            z0.im+=z1.im;
         }

         if ((t==0)&&(np>=3))
            wc=-0.5;
         else
            wc=-1.0;

         z0.re*=wc;
         z0.im*=wc;

         add_to_hsum(idsm,(double*)(&z0));
      }
   }
}


complex_dble AGn_action(int n,sl3_alg_dble *ad)
{
   complex_dble z;

   z.re=0.0;
   z.im=0.0;

   if (n<=1)
      return z;
   else if (n>ns)
      alloc_arrays(n);

   copy_bnd_ad(0,ad);
   reset_hsum(idsm);
   local_action(n,ad);
   global_hsum(idsm,(double*)(&z));

   return z;
}
