
/*******************************************************************************
*
* File Amn_action.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Calculation of the homogeneous parts of the measure action.
*
*   complex_dble Amn_action(n,sl3_alg_dble *ad)
*     Returns the contribution of degree n in the gauge potential ad
*     to the measure action.
*
* The program Amn_action() acts globally and must be called on all MPI
* processes simultaneously.
*
*******************************************************************************/

#define AMN_ACTION_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "su3fcts.h"
#include "aflds.h"
#include "vxfcts.h"
#include "global.h"

static int idsm,ns=0;
static complex_dble *trX;
static su3_dble *eX,*Xw;


static void alloc_arrays(int n)
{
   int i;

   if (ns==0)
      idsm=init_hsum(2);

   if (n>ns)
   {
      if (ns>0)
      {
         free(trX);
         afree(eX);
      }

      trX=malloc((n+1)*sizeof(*trX));
      eX=amalloc((n+2)*sizeof(*eX),ALIGN);

      error((trX==NULL)||(eX==NULL),1,"alloc_arrays [Amn_action.c]",
            "Unable to allocate auxiliary arrays");
      ns=n;
   }

   Xw=eX+n+1;

   for (i=0;i<=n;i++)
   {
      trX[i].re=0.0;
      trX[i].im=0.0;
   }

   cm3x3_zero(n+2,eX);
}


static void exp_to_poly(int n,sl3_alg_dble *X)
{
   int i;
   double r;

   cm3x3_unity(1,eX);
   sl3_to_su3(X,Xw);

   for (i=1;i<=n;i++)
   {
      su3xsu3(Xw,eX+i-1,eX+i);
      r=1.0/(double)(i);
      cm3x3_mulr(&r,eX+i,eX+i);
   }

   for (i=0;i<=n;i++)
   {
      trX[i].re=eX[i].c11.re+eX[i].c22.re+eX[i].c33.re;
      trX[i].im=eX[i].c11.im+eX[i].c22.im+eX[i].c33.im;
   }
}


static void local_action(int n,sl3_alg_dble *ad)
{
   int ix,ifc,k;
   complex_dble z0,z1;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      for (ifc=0;ifc<8;ifc++)
      {
         exp_to_poly(n,ad);

         z0.re=0.0;
         z0.im=0.0;

         for (k=0;k<=n;k++)
         {
            z1.re=trX[k].re*trX[n-k].re-trX[k].im*trX[n-k].im;
            z1.im=trX[k].re*trX[n-k].im+trX[k].im*trX[n-k].re;

            if (k&0x1)
            {
               z0.re-=z1.re;
               z0.im-=z1.im;
            }
            else
            {
               z0.re+=z1.re;
               z0.im+=z1.im;
            }
         }

         add_to_hsum(idsm,(double*)(&z0));
         ad+=1;
      }
   }
}


complex_dble Amn_action(int n,sl3_alg_dble *ad)
{
   double r;
   complex_dble z;

   z.re=0.0;
   z.im=0.0;

   if ((n<2)||(n&0x1))
      return z;
   else if (n>ns)
      alloc_arrays(n);

   reset_hsum(idsm);
   local_action(n,ad);
   global_hsum(idsm,(double*)(&z));

   r=-bernoulli(n)/(double)(n);
   z.re*=r;
   z.im*=r;

   return z;
}
