
/*******************************************************************************
*
* File divad.c
*
* Copyright (C) 2015, 2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Divergence of the gauge potential and gauge-fixing action.
*
*   void divad(sl3_alg_dble *ad,sl3_alg_dble *cd)
*     Assigns the divergence of the gauge potential ad to the ghost
*     field cd (see the notes).
*
*   complex_dble gfix_action(sl3_alg_dble *ad1,sl3_alg_dble *ad2)
*     Returns the value of the bilinear form defined by the gauge-fixing
*     action (see the notes).
*
* The layout of the arrays ad and cd, representing the gauge potential
* A_mu(x) the ghost field C(x), must be as described in main/README.global,
* README.aflds and cflds/README.cflds.
*
* For all points x in the range 0<x0<T, the field computed by divad() is
* given by
*
*  C(x)=sum_mu{A_mu(x)-A_mu(x-mu)},
*
* while at x0=0
*
*  C(x)=2*A_0(x)+sum_{mu>0}{A_mu(x)-A_mu(x-mu)}
*
* NOTE: This definition is chosen for both SF and open-SF boundary conditions.
*
* In the case of open-SF boundary conditions, the bilinear form computed by
* the program gfix_action() is
*
*  (1/2)*sum_{x0=0}sum_{x1,x2,x3}w(x0)*C1^a(x)*C2^a(x)
*
* where C1 and C2 are the divergencies of the input gauge fields A1 and A2 as
* computed by divad(). The weight factor w(x0) is 1/2 at x0=0 and 1 otherwise.
*
* If SF boundary conditions are chosen, the gauge-fixing action is given by
* the same expression, with weight w(0)=0, plus
*
*  (2*N0^3*N1*N2*N3)^(-1)*sum_{y,z}A1^a_0(x)*A2^a_0(z),
*
* where N0=NPROC0*L0,..,N3=NPROC3*L3. See the notes
*
*  M. Luescher: "Normalization of the gradient-flow coupling in LQCD"
*
* [doc/gfnorm.pdf] and
*
*  M. Luescher: "Lattice QCD vertices in ISPT"
*
* [doc/vxfcts.pdf] for further explanations.
*
* The program gfix_action() requires a workspace of 2 ghost fields. All
* programs in this module perform global communications and must be called
* simultaneously on all MPI processes.
*
*******************************************************************************/

#define DIVAD_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "pmdb.h"
#include "lattice.h"
#include "linalg.h"
#include "su3fcts.h"
#include "cflds.h"
#include "aflds.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static int idsm,idA1,idA2,init=0;


static void set_ids(void)
{
   idsm=init_hsum(2);
   idA1=init_hsum(16);
   idA2=init_hsum(16);
   init=1;
}


void divad(sl3_alg_dble *ad,sl3_alg_dble *cd)
{
   int ix,iy,t,mu;

   set_calg2zero(VOLUME/2,cd);
   set_calg2zero(BNDRY/2,cd+VOLUME);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      if (t==0)
      {
         _sl3_alg_mulr_assign(cd[ix],2.0,*ad);
      }
      else
         cd[ix]=(*ad);

      iy=iup[ix][0];

      if (t<(N0-1))
      {
         _sl3_alg_sub_assign(cd[iy],*ad);
      }

      ad+=1;

      if (t>0)
      {
         _sl3_alg_sub_assign(cd[ix],*ad);
      }

      iy=idn[ix][0];

      if (t==1)
      {
         _sl3_alg_mulr_add_assign(cd[iy],2.0,*ad);
      }
      else
      {
         _sl3_alg_add_assign(cd[iy],*ad);
      }

      ad+=1;

      for (mu=1;mu<4;mu++)
      {
         _sl3_alg_add_assign(cd[ix],*ad);
         iy=iup[ix][mu];
         _sl3_alg_sub_assign(cd[iy],*ad);
         ad+=1;

         _sl3_alg_sub_assign(cd[ix],*ad);
         iy=idn[ix][mu];
         _sl3_alg_add_assign(cd[iy],*ad);
         ad+=1;
      }
   }

   cpcd_ext_bnd(0,cd);
}


complex_dble gfix_action(sl3_alg_dble *ad1,sl3_alg_dble *ad2)
{
   int bc,ix,t;
   double r;
   complex_dble z,w;
   sl3_alg_dble **wcd;

   if (init==0)
      set_ids();

   wcd=reserve_wcd(2);
   divad(ad1,wcd[0]);
   divad(ad2,wcd[1]);
   bc=bc_type();
   reset_hsum(idsm);

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);
      z=sl3_biform(wcd[0]+ix,wcd[1]+ix);

      if (t>0)
         add_to_hsum(idsm,(double*)(&z));
      else if (bc==2)
      {
         z.re*=0.5;
         z.im*=0.5;
         add_to_hsum(idsm,(double*)(&z));
      }
   }

   global_hsum(idsm,(double*)(&z));

   if (bc==1)
   {
      reset_hsum(idA1);
      reset_hsum(idA2);

      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         add_to_hsum(idA1,(double*)(ad1));
         add_to_hsum(idA2,(double*)(ad2));
         ad1+=1;
         ad2+=1;
         add_to_hsum(idA1,(double*)(ad1));
         add_to_hsum(idA2,(double*)(ad2));
         ad1+=7;
         ad2+=7;
      }

      global_hsum(idA1,(double*)(wcd[0]));
      global_hsum(idA2,(double*)(wcd[1]));

      w=sl3_biform(wcd[0],wcd[1]);
      r=1.0/((double)(N0*N0*N0)*(double)(N1*N2*N3));
      z.re+=r*w.re;
      z.im+=r*w.im;
   }

   release_wcd();
   z.re*=0.5;
   z.im*=0.5;

   return z;
}
