
/*******************************************************************************
*
* File Eclov.c
*
* Copyright (C) 2015, 2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Perturbative expansion of the Yang-Mills action density.
*
*   void Eclov(int n,su3_dble *ud,complex_dble *E_sp,complex_dble *E_tm)
*     Computes the Yang-Mills action density up to order n in the gauge
*     coupling, assuming ud contains the link fields obtained in stochastic
*     perturbation theory up to order n (see the notes). The average of
*     the spatial and time-like components of the density are calculated at
*     global time x0=T/2 (where T=NPROC0*L0) and their coefficients at order
*     k=0,..,n are assigned to E_sp[k] and E_tm[k], respectively.
*
*   void Sclov(int n,su3_dble *ud,complex_dble *S)
*     Computes the Yang-Mills gauge action up to order n in the gauge coupling,
*     assuming ud contains the link fields obtained in stochastic perturbation
*     theory up to order n (see the notes).
*
* The Yang-Mills action density s(x) is defined by
*
*  s(x)=1/(4*g0^2)*sum_{mu,nu} [F_{mu,nu}^a(x)]^2
*
* where
*
*  F_{mu,nu}^a(x)=-2*tr{F_{mu,nu}(x)*T^a}, a=1,..,8,
*
* are the SL(3,C) components of the symmetric field tensor returned by the
* program ftensor() [ftensor.c]. At the boundaries of the lattice at global
* time 0 and T=NPROC0*L0, the action density is set to zero. The action
* computed by Sclov() is the sum over all points x of the action density.
*
* The program Eclov() returns the contributions to the action density from
* the spatial planes (mu,nu>0) and the time-like planes (mu=0,nu>0 or nu=0,
* mu>0) separately. Their sum is equal to the average of s(x) at x0=T/2.
*
* The array ud is expected to contain the n+1 coefficient fields in the
* expansion
*
*   U=1+g0*U[0]+g0^2*U[1]+..+g0^(n+1)*U[n]
*
* of the gauge field in powers of the coupling g0 (cf. uflds/uflds.c).
*
* The programs in this module perform global operations and must be called
* simultaneously on all MPI processes.
*
*******************************************************************************/

#define ECLOV_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "utils.h"
#include "aflds.h"
#include "uflds.h"
#include "su3fcts.h"
#include "lattice.h"
#include "ftensor.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)
#define NMAX 16

static int ms=0;
static int nsm[NMAX],ism[NMAX][2],isp,itm;
static complex_dble *clov;
static su3_dble *Xw;


static void alloc_arrays(int n)
{
   int m,i;

   if (ms==0)
   {
      for (i=0;i<NMAX;i++)
      {
         nsm[i]=0;
         ism[i][0]=0;
         ism[i][1]=0;
      }
   }

   if (ms>0)
   {
      free(clov);
      afree(Xw);
   }

   m=n+1;
   clov=malloc(m*sizeof(*clov));
   Xw=amalloc(m*sizeof(*Xw),ALIGN);

   error((clov==NULL)||(Xw==NULL),1,
         "alloc_arrays [Eclov.c]","Unable to allocate auxiliary arrays");

   for (i=0;i<m;i++)
   {
      clov[i].re=0.0;
      clov[i].im=0.0;
   }

   cm3x3_zero(m,Xw);
   ms=m;
}


static void set_ism(int n)
{
   int m,i;

   m=2*(n+1);

   for (i=0;i<NMAX;i++)
   {
      if (nsm[i]==0)
      {
         nsm[i]=m;
         ism[i][0]=init_hsum(m);
         ism[i][1]=init_hsum(m);
      }

      if (nsm[i]==m)
      {
         isp=ism[i][0];
         itm=ism[i][1];
         return;
      }
   }

   error_loc(1,1,"set_ism [Eclov.c]",
             "Requested too many hsum instances");
}


static void rm_trace(int n,su3_dble *ft)
{
   complex_dble z;
   su3_dble *ftm;

   ftm=ft+n+1;

   for (;ft<ftm;ft++)
   {
      z.re=(1.0/3.0)*((*ft).c11.re+(*ft).c22.re+(*ft).c33.re);
      z.im=(1.0/3.0)*((*ft).c11.im+(*ft).c22.im+(*ft).c33.im);

      (*ft).c11.re-=z.re;
      (*ft).c11.im-=z.im;
      (*ft).c22.re-=z.re;
      (*ft).c22.im-=z.im;
      (*ft).c33.re-=z.re;
      (*ft).c33.im-=z.im;
   }
}


static void local_Eclov(int n,su3_dble *ud)
{
   int m,i;
   int ix,t,pln;
   su3_dble **ft,*ftn;

   m=n+1;
   ft=ftensor(n,ud);

   for (pln=0;pln<6;pln++)
   {
      ftn=ft[pln];

      for (ix=0;ix<VOLUME;ix++)
      {
         t=global_time(ix);

         if (t==(N0/2))
         {
            rm_trace(n,ftn);
            cm3x3_mul_poly(n,ftn,ftn,Xw);

            for (i=0;i<m;i++)
            {
               clov[i].re=-Xw[i].c11.re-Xw[i].c22.re-Xw[i].c33.re;
               clov[i].im=-Xw[i].c11.im-Xw[i].c22.im-Xw[i].c33.im;
            }

            if (pln>=3)
               add_to_hsum(isp,(double*)(clov));
            else
               add_to_hsum(itm,(double*)(clov));
         }

         ftn+=m;
      }
   }
}


void Eclov(int n,su3_dble *ud,complex_dble *E_sp,complex_dble *E_tm)
{
   int i;
   double r;

   if ((n+1)>ms)
      alloc_arrays(n);

   set_ism(n);
   reset_hsum(isp);
   reset_hsum(itm);
   local_Eclov(n,ud);
   global_hsum(isp,(double*)(E_sp));
   global_hsum(itm,(double*)(E_tm));
   r=1.0/(double)(N1*N2*N3);

   for (i=0;i<=n;i++)
   {
      E_sp[i].re*=r;
      E_sp[i].im*=r;
      E_tm[i].re*=r;
      E_tm[i].im*=r;
   }
}


static void local_Sclov(int n,su3_dble *ud)
{
   int m,i;
   int ix,t,pln;
   su3_dble **ft,*ftn;

   m=n+1;
   ft=ftensor(n,ud);

   for (pln=0;pln<6;pln++)
   {
      ftn=ft[pln];

      for (ix=0;ix<VOLUME;ix++)
      {
         t=global_time(ix);

         if (t!=0)
         {
            rm_trace(n,ftn);
            cm3x3_mul_poly(n,ftn,ftn,Xw);

            for (i=0;i<m;i++)
            {
               clov[i].re=-Xw[i].c11.re-Xw[i].c22.re-Xw[i].c33.re;
               clov[i].im=-Xw[i].c11.im-Xw[i].c22.im-Xw[i].c33.im;
            }

            add_to_hsum(isp,(double*)(clov));
         }

         ftn+=m;
      }
   }
}


void Sclov(int n,su3_dble *ud,complex_dble *S)
{
   if ((n+1)>ms)
      alloc_arrays(n);

   set_ism(n);
   reset_hsum(isp);
   local_Sclov(n,ud);
   global_hsum(isp,(double*)(S));
}
