
/*******************************************************************************
*
* File valrt.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of the values of rooted labeled tree diagrams.
*
*   void valrt(rtree_t *rt,void *vrt)
*     Evaluates the diagram "rt" using the random fields set by the last
*     call of generate_rflds() [rflds.c]. The value of the diagram *with
*     amputated root line* is assigned to the field "vrt".
*
*   void valrt_wspace(rtree_t *rt,int *nw)
*     Determines the numbers nw[0],..,nw[3] of scalar, gluon, ghost and
*     quark fields reserved by the workspace facility, when the diagram
*     "rt" is evaluated by the program valrt().
*
* The programs in this module obtain the theory and ISPT parameters from the
* parameter data base. It is thus taken for granted that the parameters have
* been set through set_th_parms() and set_ispt_parms().

* Currently the program valrt() can only evaluate the diagrams in the YM
* and LYM theories.
*
* The program valrt() performs global operations and must be simultaneously
* called on all MPI processes.
*
*******************************************************************************/

#define VALRT_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "pmdb.h"
#include "aflds.h"
#include "cflds.h"
#include "vxfcts.h"
#include "ispt.h"

static int init=0,order,nwmx[4],nwloc[4];
static sl3_alg_dble **aflds,**asrc,*was,*wcs;
static theory_t th;
static vlist_t vls;


static void set_th(void)
{
   th_parms_t thp;
   ispt_parms_t isp;

   thp=th_parms();
   th=thp.th;
   isp=ispt_parms();
   order=isp.order;

   error_loc(th==THEORIES,1,"set_th [valrt.c]",
             "Theory parameters are not set");
   error_loc(order==0,1,"set_th [valrt.c]",
             "ISPT parameters are not set");

   vls=vertices(th,order);
   init=1;
}


static void eval_vx(rtree_t *rt,rflds_t *rf)
{
   int nsrc,nbr,*v;
   int na,ia,isrc,i;
   ltype_t lt,ltn;
   vertex_t *vx;
   src_t *src;
   rtree_t *br;
   sl3_alg_dble **a,*cb,*c;
   sl3_alg_dble **wad,**wcd;

   nsrc=(*rt).nsrc;
   nbr=(*rt).nbr;
   v=(*rt).v;
   lt=(*rt).lt;
   src=(*rt).src;
   br=(*rt).br;

   vx=vls.v[v[0]]+v[1];
   na=(*vx).nl[(int)(GLUON)];
   if (lt==GLUON)
      na-=1;
   ia=0;
   a=asrc;
   c=NULL;
   cb=NULL;

   for (i=0;i<nbr;i++)
   {
      asrc=a+na;
      ltn=br[i].lt;

      if (ltn==GLUON)
      {
         eval_vx(br+i,rf);
         wad=reserve_wad(1);
         aprop(was,wad[0]);
         a[ia]=wad[0];
         ia+=1;
      }
      else if (ltn==GHOST)
      {
         eval_vx(br+i,rf);
         wcd=reserve_wcd(1);
         cprop(wcs,wcd[0]);
         c=wcd[0];
      }
      else if (ltn==ANTIGHOST)
      {
         eval_vx(br+i,rf);
         wcd=reserve_wcd(1);
         cprop(wcs,wcd[0]);
         cb=wcd[0];
      }
   }

   for (i=0;i<nsrc;i++)
   {
      ltn=src[i].lt;
      isrc=src[i].isrc;

      if (ltn==GLUON)
      {
         a[ia]=(sl3_alg_dble*)((*rf).flds[(int)(ltn)][isrc]);
         ia+=1;
      }
      else if (ltn==GHOST)
         c=(sl3_alg_dble*)((*rf).flds[(int)(ltn)][isrc]);
      else if (ltn==ANTIGHOST)
         cb=(sl3_alg_dble*)((*rf).flds[(int)(ltn)][isrc]);
   }

   if (lt==GLUON)
      na+=1;

   if ((*vx).type==VG)
      AGn(na,a,was);
   else if ((*vx).type==VGM)
      Amn(na,a,was);
   else if ((*vx).type==VFP)
   {
      if (lt==GLUON)
         AFPn(na,a,cb,c,was);
      else if (lt==GHOST)
         CFPn(na,a,c,wcs);
      else if (lt==ANTIGHOST)
         CbFPn(na,a,cb,wcs);
      else
         error_root(1,1,"eval_vx [valrt.c]","Unknown vertex function");
   }
   else if ((*vx).type==VGF)
      Agf(a[0],was);
   else
      error_root(1,2,"eval_vx [valrt.c]","Unknown vertex function");

   for (i=0;i<nbr;i++)
   {
      ltn=br[i].lt;

      if (ltn==GLUON)
         release_wad();
      else if ((ltn==GHOST)||(ltn==ANTIGHOST))
         release_wcd();
   }
}


void valrt(rtree_t *rt,void *vrt)
{
   sl3_alg_dble **wcd;
   rflds_t *rf;

   if (init==0)
      set_th();

   rf=rflds();

   if ((th==YM)||(th==LYM))
   {
      if (init==1)
      {
         aflds=malloc(2*order*sizeof(*aflds));
         error(aflds==NULL,1,"valrt [valrt.c]",
               "Unable to allocate auxiliary array");
         init=2;
      }

      was=(sl3_alg_dble*)(vrt);
      wcd=reserve_wcd(1);
      wcs=wcd[0];
      asrc=aflds;

      eval_vx(rt,rf);

      release_wcd();
   }
}


static void eval_wspace(rtree_t *rt)
{
   int i,nbr,*v;
   ltype_t lt,ltn;
   vertex_t *vx;
   rtree_t *br;

   nbr=(*rt).nbr;
   v=(*rt).v;
   lt=(*rt).lt;
   br=(*rt).br;
   vx=vls.v[v[0]]+v[1];

   for (i=0;i<nbr;i++)
   {
      eval_wspace(br+i);
      ltn=br[i].lt;

      if (ltn==SCALAR)
         nwloc[0]+=1;
      else if (ltn==GLUON)
         nwloc[1]+=1;
      else if ((ltn==GHOST)||(ltn==ANTIGHOST))
         nwloc[2]+=1;
      else
         nwloc[3]+=1;
   }

   if (((*vx).type==VFP)&&((lt==GHOST)||(lt==ANTIGHOST)))
      nwloc[1]+=1;

   if ((*vx).type==VGF)
      nwloc[2]+=1;

   for (i=0;i<4;i++)
   {
      if (nwloc[i]>nwmx[i])
         nwmx[i]=nwloc[i];
   }

   if (((*vx).type==VFP)&&((lt==GHOST)||(lt==ANTIGHOST)))
      nwloc[1]-=1;

   if ((*vx).type==VGF)
      nwloc[2]-=1;

   for (i=0;i<nbr;i++)
   {
      ltn=br[i].lt;

      if (ltn==SCALAR)
         nwloc[0]-=1;
      else if (ltn==GLUON)
         nwloc[1]-=1;
      else if ((ltn==GHOST)||(ltn==ANTIGHOST))
         nwloc[2]-=1;
      else
         nwloc[3]-=1;
   }
}


void valrt_wspace(rtree_t *rt,int *nw)
{
   int i;

   if (init==0)
      set_th();

   for (i=0;i<4;i++)
   {
      nwmx[i]=0;
      nwloc[i]=0;
   }

   if (th!=PHI4)
   {
      nwmx[2]=1;
      nwloc[2]=1;

      if ((th==QCD)||(th==LQCD))
      {
         nwmx[3]=1;
         nwloc[3]=1;
      }
   }

   eval_wspace(rt);

   for (i=0;i<4;i++)
      nw[i]=nwmx[i];
}
