
/*******************************************************************************
*
* File bcnds.c
*
* Copyright (C) 2005, 2010-2014, 2015 Martin Luescher, John Bulava
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Programs related to the boundary conditions in the time direction.
*
*   int *bnd_pts(int *n)
*     Returns the address of an array of length n whose elements are the
*     indices of the points on the local lattice at global time 0. The
*     ordering of the indices is such that the n/2 even points come first.
*
*   void avg_cd0(sl3_alg_dble *cd,sl3_alg_dble *ct0)
*     On the MPI processes where cpr[0]=0, this program computes the
*     average ct0 of the field cd at global time 0. Elsewhere ct0 is set
*     to zero.
*
*   void set_adbc(sl3_alg_dble *ad)
*     Sets the field variables at time 0 and T of the gauge potential with
*     address ad to zero according to the boundary conditions specified in
*     the parameter data base.
*
*   int check_adbc(sl3_alg_dble *ad)
*     Returns 1 if the gauge potential with address ad satisfies the
*     boundary conditions specified in the parameter data base. Otherwise
*     0 is returned.
*
*   void set_cdbc(sl3_alg_dble *cd)
*     Sets the field variables at time 0 of the ghost field with address
*     cd according to the boundary conditions specified in the parameter
*     data base. In particular, in the case of SF boundary conditions, the
*     field at global time 0 is replaced by its average value at this time.
*
*   int check_cdbc(sl3_alg_dble *cd)
*     Returns 1 if the ghost field with address cd satisfies the boundary
*     conditions specified in the parameter data base. Otherwise 0 is
*     returned.
*
*   void set_sdbc(spinor_dble *sd)
*     Sets the field variables at time 0 of the quark field with address sd
*     to zero.
*
*   int check_sdbc(spinor_dble *sd)
*     Returns 1 if the quark field with address sd vanishes at time 0.
*     Otherwise 0 is returned.
*
* The time extent T of the lattice is NPROC0*L0 for both SF and open-SF
* boundary conditions. Note that the points at time T are not contained in
* the local lattices.
*
* See main/README.global, aflds/README.aflds and cflds/README.cflds for a
* description of the layout of the fields and the boundary conditions.
*
* The programs in this module may perform global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define BCNDS_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)

static int init=0,idsm,root,npts,*pts;
static sl3_alg_dble ad0={{0.0,0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0},
                         {0.0,0.0},{0.0,0.0},{0.0,0.0},{0.0,0.0}};
static spinor_dble sd0={{{0.0,0.0},{0.0,0.0},{0.0,0.0}},
                        {{0.0,0.0},{0.0,0.0},{0.0,0.0}},
                        {{0.0,0.0},{0.0,0.0},{0.0,0.0}},
                        {{0.0,0.0},{0.0,0.0},{0.0,0.0}}};
static MPI_Comm MPI_COMM_TIME0;


static void set_ids(void)
{
   idsm=init_hsum(16);
}


static void set_communicator(void)
{
   int ip,ip1,ip2,ip3;
   int cp[4],ranks[NPROC/NPROC0];
   MPI_Group world,time0;

   ip=0;
   cp[0]=0;

   for (ip1=0;ip1<NPROC1;ip1++)
   {
      for (ip2=0;ip2<NPROC2;ip2++)
      {
         for (ip3=0;ip3<NPROC3;ip3++)
         {
            cp[1]=ip1;
            cp[2]=ip2;
            cp[3]=ip3;

            ranks[ip]=ipr_global(cp);
            ip+=1;
         }
      }
   }

   MPI_Comm_group(MPI_COMM_WORLD,&world);
   MPI_Group_incl(world,NPROC/NPROC0,ranks,&time0);
   MPI_Comm_create(MPI_COMM_WORLD,time0,&MPI_COMM_TIME0);
   MPI_Group_translate_ranks(world,1,ranks,time0,&root);
}


static void alloc_pts(void)
{
   int ix,*pt;

   error(iup[0][0]==0,1,"alloc_pts [bcnds.c]","Geometry arrays are not set");

   if (cpr[0]==0)
   {
      npts=L1*L2*L3;
      pts=malloc(npts*sizeof(*pts));

      if (pts!=NULL)
      {
         pt=pts;

         for (ix=0;ix<VOLUME;ix++)
         {
            if (global_time(ix)==0)
            {
               (*pt)=ix;
               pt+=1;
            }
         }
      }
   }
   else
   {
      npts=0;
      pts=NULL;
   }

   error((npts>0)&&(pts==NULL),1,"alloc_pts [bcnds.c]",
         "Unable to allocate index array");

   set_ids();
   set_communicator();
   init=1;
}


int *bnd_pts(int *n)
{
   if (init==0)
      alloc_pts();

   (*n)=npts;

   return pts;
}


void avg_cd0(sl3_alg_dble *cd,sl3_alg_dble *ct0)
{
   int *pt,*ptm;
   double r;
   sl3_alg_dble sm;

   if (init==0)
      alloc_pts();

   if (cpr[0]==0)
   {
      reset_hsum(idsm);
      pt=pts;
      ptm=pts+npts;

      for (;pt<ptm;pt++)
         add_to_hsum(idsm,(double*)(cd+(*pt)));

      if ((NPROC/NPROC0)>1)
      {
         local_hsum(idsm,(double*)(&sm));
         MPI_Reduce(&sm,ct0,16,MPI_DOUBLE,MPI_SUM,root,MPI_COMM_TIME0);
         MPI_Bcast(ct0,16,MPI_DOUBLE,root,MPI_COMM_TIME0);
      }
      else
         local_hsum(idsm,(double*)(ct0));

      r=1.0/(double)(N1*N2*N3);
      _sl3_alg_mulr_assign(*ct0,r,*ct0);
   }
   else
      (*ct0)=ad0;
}


void set_adbc(sl3_alg_dble *ad)
{
   int k,*pt,*ptm;
   sl3_alg_dble *X;

   if (init==0)
      alloc_pts();

   if ((cpr[0]==0)&&(bc_type()==1))
   {
      pt=pts+(npts/2);
      ptm=pts+npts;

      for (;pt<ptm;pt++)
      {
         X=ad+8*(pt[0]-(VOLUME/2));

         for (k=2;k<8;k++)
            X[k]=ad0;
      }
   }

   if (cpr[0]==(NPROC0-1))
   {
      X=ad+4*VOLUME+7*(BNDRY/4);

      for (k=0;k<3;k++)
         X[k]=ad0;
   }
}


void set_cdbc(sl3_alg_dble *cd)
{
   int *pt,*pm;
   sl3_alg_dble ct0;

   if (bc_type()==1)
   {
      avg_cd0(cd,&ct0);

      if (cpr[0]==0)
      {
         pt=pts;
         pm=pts+npts;

         for (;pt<pm;pt++)
            cd[*pt]=ct0;
      }
   }
}


void set_sdbc(spinor_dble *sd)
{
   int *pt,*pm;

   if (init==0)
      alloc_pts();

   if (cpr[0]==0)
   {
      pt=pts;
      pm=pts+npts;

      for (;pt<pm;pt++)
         sd[*pt]=sd0;
   }
}


static int is_equal_ad(sl3_alg_dble *X,sl3_alg_dble *Y)
{
   int it;

   it=1;
   it&=((*X).c1.re==(*Y).c1.re);
   it&=((*X).c1.im==(*Y).c1.im);
   it&=((*X).c2.re==(*Y).c2.re);
   it&=((*X).c2.im==(*Y).c2.im);
   it&=((*X).c3.re==(*Y).c3.re);
   it&=((*X).c3.im==(*Y).c3.im);
   it&=((*X).c4.re==(*Y).c4.re);
   it&=((*X).c4.im==(*Y).c4.im);
   it&=((*X).c5.re==(*Y).c5.re);
   it&=((*X).c5.im==(*Y).c5.im);
   it&=((*X).c6.re==(*Y).c6.re);
   it&=((*X).c6.im==(*Y).c6.im);
   it&=((*X).c7.re==(*Y).c7.re);
   it&=((*X).c7.im==(*Y).c7.im);
   it&=((*X).c8.re==(*Y).c8.re);
   it&=((*X).c8.im==(*Y).c8.im);

   return it;
}


static int is_zero_sd(spinor_dble *X)
{
   int it;

   it=1;
   it&=((*X).c1.c1.re==0.0);
   it&=((*X).c1.c1.im==0.0);
   it&=((*X).c1.c2.re==0.0);
   it&=((*X).c1.c2.im==0.0);
   it&=((*X).c1.c3.re==0.0);
   it&=((*X).c1.c3.im==0.0);

   it&=((*X).c2.c1.re==0.0);
   it&=((*X).c2.c1.im==0.0);
   it&=((*X).c2.c2.re==0.0);
   it&=((*X).c2.c2.im==0.0);
   it&=((*X).c2.c3.re==0.0);
   it&=((*X).c2.c3.im==0.0);

   it&=((*X).c3.c1.re==0.0);
   it&=((*X).c3.c1.im==0.0);
   it&=((*X).c3.c2.re==0.0);
   it&=((*X).c3.c2.im==0.0);
   it&=((*X).c3.c3.re==0.0);
   it&=((*X).c3.c3.im==0.0);

   it&=((*X).c4.c1.re==0.0);
   it&=((*X).c4.c1.im==0.0);
   it&=((*X).c4.c2.re==0.0);
   it&=((*X).c4.c2.im==0.0);
   it&=((*X).c4.c3.re==0.0);
   it&=((*X).c4.c3.im==0.0);

   return it;
}


int check_adbc(sl3_alg_dble *ad)
{
   int it,is;
   int k,*pt,*ptm;
   sl3_alg_dble *X;

   if (init==0)
      alloc_pts();

   it=1;

   if ((cpr[0]==0)&&(bc_type()==1))
   {
      pt=pts+(npts/2);
      ptm=pts+npts;

      for (;pt<ptm;pt++)
      {
         X=ad+8*(pt[0]-(VOLUME/2));

         for (k=2;k<8;k++)
            it&=is_equal_ad(X+k,&ad0);
      }
   }

   if (cpr[0]==(NPROC0-1))
   {
      X=ad+4*VOLUME+7*(BNDRY/4);

      for (k=0;k<3;k++)
         it&=is_equal_ad(X+k,&ad0);
   }

   if (NPROC>1)
   {
      is=it;
      MPI_Allreduce(&is,&it,1,MPI_INT,MPI_MIN,MPI_COMM_WORLD);
   }

   return it;
}


int check_cdbc(sl3_alg_dble *cd)
{
   int it,is;
   int *pt,*pm;
   sl3_alg_dble ct0;

   if (init==0)
      alloc_pts();

   it=1;

   if (bc_type()==1)
   {
      if (cpr[0]==0)
      {
         ct0=cd[pts[0]];
         MPI_Bcast(&ct0,16,MPI_DOUBLE,root,MPI_COMM_TIME0);

         pt=pts;
         pm=pts+npts;

         for (;pt<pm;pt++)
            it&=is_equal_ad(cd+(*pt),&ct0);
      }

      if (NPROC>1)
      {
         is=it;
         MPI_Allreduce(&is,&it,1,MPI_INT,MPI_MIN,MPI_COMM_WORLD);
      }
   }

   return it;
}


int check_sdbc(spinor_dble *sd)
{
   int it,is;
   int *pt,*pm;

   if (init==0)
      alloc_pts();

   it=1;

   if (cpr[0]==0)
   {
      pt=pts;
      pm=pts+npts;

      for (;pt<pm;pt++)
         it&=is_zero_sd(sd+(*pt));
   }

   if (NPROC>1)
   {
      is=it;
      MPI_Allreduce(&is,&it,1,MPI_INT,MPI_MIN,MPI_COMM_WORLD);
   }

   return it;
}
