
/*******************************************************************************
*
* File diagram.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Utility programs for diagrams.
*
* The externally accessible functions are
*
*   diagram_t *alloc_dgm(int nv,int nl)
*     Returns the pointer to a newly allocated diagram structure with
*     properly allocated and initialized arrays (integers to 0, line
*     types to LTYPES).
*
*   void free_dgm(diagram_t *dm)
*     Frees the arrays in the diagram structure dm and then the structure
*     itself.
*
*   void copy_dgm(diagram_t *dm1,diagram_t *dm2)
*     Copies the diagram dm1 to dm2.
*
*   void remove_line(int i,diagram_t *dm1,diagram_t *dm2)
*     Obtains a new diagram dm2 by removing the line number i from
*     the diagram dm1.
*
*   void connected_part(int i,diagram_t *dm1,diagram_t *dm2)
*     Finds the connected subdiagram dm2 of dm1 that contains the
*     vertex number i.
*
* Notes:
*
* Diagrams are described by structures of type diagram_t containing a
* list of vertices and lines. The elements of the structure are
*
*  nv                    Number of vertices.
*
*  nl                    Number of internal lines.
*
*  v[i][0],v[i][1]       Indices of the vertex number i in the list returned
*                        by the program vertices() [vertices.c] (i=0,..,nv-1).
*
*  l[j]                  Structure of type line_t describing the internal
*                        line number j (j=0,..,nl-1).
*
* The elements of the line_t structure are
*
*  type                  Type of the line (SCALAR,GLUON,..).
*
*  iv[0],iv[1]           Indices of the vertices in the diagram structure
*                        where the line starts and ends.
*
* Diagrams must have at least one vertex but may have no internal lines.
*
* External lines are not included in the diagram structure. Their number and
* type can however be infered from the list of vertices and internal lines.
* It is also possible to effectively include them by adding the appropriate
* 1-point vertices at their outer end. These lines then turn into internal
* ones connecting the 1-point vertices with the rest of the diagram. Single-
* line diagrams, for example, are represented by a pair of 1-point vertices
* and the connecting line.
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define DIAGRAM_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "diagrams.h"

static int nvs=0,nls=0;
static int *ivs,*ils;


diagram_t *alloc_dgm(int nv,int nl)
{
   int i,(*v)[2];
   line_t *l;
   diagram_t *dm;

   error_loc((nv<1)||(nl<0),1,"alloc_dgm [diagram.c]",
             "Arguments are out of range");

   dm=malloc(1*sizeof(*dm));
   v=malloc(nv*sizeof(*v));

   if (nl>0)
      l=malloc(nl*sizeof(*l));
   else
      l=NULL;

   error_loc((dm==NULL)||(v==NULL)||((nl>0)&&(l==NULL)),1,
             "alloc_dgm [diagram.c]","Unable to allocate data arrays");

   (*dm).nv=nv;
   (*dm).nl=nl;
   (*dm).v=v;
   (*dm).l=l;

   for (i=0;i<nv;i++)
   {
      (*v)[0]=0;
      (*v)[1]=0;
      v+=1;
   }

   for (i=0;i<nl;i++)
   {
      (*l).type=LTYPES;
      (*l).iv[0]=0;
      (*l).iv[1]=0;
      l+=1;
   }

   return dm;
}


void free_dgm(diagram_t *dm)
{
   if (dm!=NULL)
   {
      free((*dm).v);

      if ((*dm).nl>0)
         free((*dm).l);

      free(dm);
   }
}


void copy_dgm(diagram_t *dm1,diagram_t *dm2)
{
   int nv,nl,i;
   int (*v1)[2],(*v2)[2];
   line_t *l1,*l2;

   nv=(*dm1).nv;
   nl=(*dm1).nl;

   error_loc((nv>(*dm2).nv)||(nl>(*dm2).nl),1,"copy_dgm [diagram.c]",
             "Improper arguments");

   (*dm2).nv=nv;
   (*dm2).nl=nl;

   v1=(*dm1).v;
   v2=(*dm2).v;

   for (i=0;i<nv;i++)
   {
      v2[0][0]=v1[0][0];
      v2[0][1]=v1[0][1];

      v1+=1;
      v2+=1;
   }

   l1=(*dm1).l;
   l2=(*dm2).l;

   for (i=0;i<nl;i++)
   {
      l2[0].type=l1[0].type;
      l2[0].iv[0]=l1[0].iv[0];
      l2[0].iv[1]=l1[0].iv[1];

      l1+=1;
      l2+=1;
   }
}


void remove_line(int i,diagram_t *dm1,diagram_t *dm2)
{
   int nv,nl,j;
   int (*v1)[2],(*v2)[2];
   line_t *l1,*l2;

   nv=(*dm1).nv;
   nl=(*dm1).nl;

   error_loc((i<0)||(i>=nl)||(nv>(*dm2).nv)||((nl-1)>(*dm2).nl),1,
             "remove_line [diagram.c]","Improper arguments");

   (*dm2).nv=nv;
   (*dm2).nl=nl-1;

   v1=(*dm1).v;
   v2=(*dm2).v;

   for (j=0;j<nv;j++)
   {
      v2[0][0]=v1[0][0];
      v2[0][1]=v1[0][1];

      v1+=1;
      v2+=1;
   }

   l1=(*dm1).l;
   l2=(*dm2).l;

   for (j=0;j<nl;j++)
   {
      if (j!=i)
      {
         l2[0].type=l1[0].type;
         l2[0].iv[0]=l1[0].iv[0];
         l2[0].iv[1]=l1[0].iv[1];
         l2+=1;
      }

      l1+=1;
   }
}


static void alloc_aux(int nv,int nl)
{
   error_loc((nv<1)||(nl<0),1,"alloc_aux [diagram.c]",
             "Arguments are out of range");

   if (nvs>0)
      free(ivs);

   if (nv<nvs)
      nv=nvs;
   if (nl<nls)
      nl=nls;

   ivs=malloc((nv+nl)*sizeof(*ivs));
   error_loc(ivs==NULL,1,"alloc_aux [diagram.c]",
             "Unable to allocate auxiliary arrays");

   nvs=nv;
   nls=nl;
   ils=ivs+nv;
}


void connected_part(int i,diagram_t *dm1,diagram_t *dm2)
{
   int nv,nl,nvn,nln;
   int s,j,k,iv1,iv2;
   int (*v1)[2],(*v2)[2];
   line_t *l1,*l2;

   nv=(*dm1).nv;
   nl=(*dm1).nl;

   error_loc((i<0)||(i>=nv),1,"connected_part [diagram.c]",
             "Improper arguments");

   if ((nv>nvs)||(nl>nls))
      alloc_aux(nv,nl);

   for (j=0;j<nv;j++)
      ivs[j]=0;

   for (j=0;j<nl;j++)
      ils[j]=0;

   ivs[i]=1;
   nvn=1;
   nln=0;
   s=1;

   while (s==1)
   {
      s=0;
      l1=(*dm1).l;

      for (j=0;j<nl;j++)
      {
         if (ils[j]==0)
         {
            iv1=l1[0].iv[0];
            iv2=l1[0].iv[1];
            k=ivs[iv1]+ivs[iv2];

            if (k==1)
            {
               ivs[iv1]=1;
               ivs[iv2]=1;
               nvn+=1;
            }

            if (k>0)
            {
               ils[j]=1;
               nln+=1;
               s=1;
            }
         }

         l1+=1;
      }
   }

   error_loc((nvn>(*dm2).nv)||(nln>(*dm2).nl),2,"connected_part [diagram.c]",
             "Improper arguments");

   (*dm2).nv=nvn;
   (*dm2).nl=nln;
   v1=(*dm1).v;
   v2=(*dm2).v;
   k=0;

   for (j=0;j<nv;j++)
   {
      if (ivs[j]==1)
      {
         v2[0][0]=v1[0][0];
         v2[0][1]=v1[0][1];
         ivs[j]=k;
         v2+=1;
         k+=1;
      }

      v1+=1;
   }

   l1=(*dm1).l;
   l2=(*dm2).l;

   for (j=0;j<nl;j++)
   {
      if (ils[j]==1)
      {
         l2[0].type=l1[0].type;
         l2[0].iv[0]=ivs[l1[0].iv[0]];
         l2[0].iv[1]=ivs[l1[0].iv[1]];
         l2+=1;
      }

      l1+=1;
   }
}
