
/*******************************************************************************
*
* File dmprop.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Diagram properties.
*
*   int check_dgm(theory_t th,int order,diagram_t *dm)
*     Checks whether the diagram dm is a possible one in the theory
*     with label th at or below the specified order in the coupling.
*     The program returns the number of external lines of the diagram
*     if it is a consistent one. Otherwise the program aborts with an
*     error message.
*
*   int is_tree(diagram_t *dm)
*     A tree diagram is a connected diagram without loops and 1-point
*     vertices. This program returns 1 if "dm" is a tree diagram and 0
*     otherwise.
*
*   int is_branch(diagram_t *dm)
*     A branch diagram is a tree diagram with only one branch. This
*     program return 1 if "dm" is a branch diagram and 0 otherwise.

*   int ncparts(diagram_t *dm)
*     Returns the number of connected parts of the diagram dm.
*
*   int nquark_loops(diagram_t *dm)
*     Returns the number of quark loops in the diagram dm.
*
*   int nghost_loops(diagram_t *dm)
*     Returns the number of ghost loops in the diagram dm.
*
*   int nlext(diagram_t *dm,vlist_t *vl,int *n1pt)
*     Returns the number of external lines of the diagram dm, assuming vl
*     points to the vertex list in the theory considered (see vertices.c).
*     The number of 1-point vertices in the diagram is assigned to n1pt.
*
*   int has_tadpole(diagram_t *dm,ltype_t lt,vlist_t *vl)
*     Returns 1 if the diagram dm has a tadpole subdiagram connected to
*     the rest of the diagram through a line of type lt (see the notes).
*     Otherwise 0 is returned.
*
*   int nVX(diagram_t *dm,vtype_t VX,vlist_t *vl)
*     Returns the number of vertices of type VX in the diagram dm, assuming
*     vl points to the list of vertices in the theory considered.
*
* A tadpole diagram is a connected diagram with one external line, one or
* more vertices and no one-point vertices.
*
* Note that the number of loops of a diagram is equal to nl-nv+nc, where nl,
* nv and nc are the numbers of internal lines, vertices and connected parts
* of the diagram, respectively.
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define DMPROP_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "rvt.h"

static int nvs=0,nls=0;
static int *ivs,*ils,**nla;
static diagram_t *dmw[2];


static void alloc_aux(int nv,int nl)
{
   int nt,i;

   error_loc((nv<1)||(nl<0),1,"alloc_aux [dmprop.c]",
             "Arguments are out of range");

   if (nvs>0)
   {
      free(ivs);
      free(nla);

      for (i=0;i<2;i++)
         free_dgm(dmw[i]);
   }

   if (nv<nvs)
      nv=nvs;
   if (nl<nls)
      nl=nls;

   nt=(int)(LTYPES);
   ivs=malloc((nv+nl+nv*nt)*sizeof(*ivs));
   nla=malloc(nv*sizeof(*nla));
   error_loc((ivs==NULL)||(nla==NULL),1,"alloc_aux [dmprop.c]",
             "Unable to allocate auxiliary arrays");

   for (i=0;i<2;i++)
      dmw[i]=alloc_dgm(nv,nl);

   nvs=nv;
   nls=nl;
   ils=ivs+nv;
   nla[0]=ils+nl;

   for (i=1;i<nv;i++)
      nla[i]=nla[i-1]+nt;
}


int check_dgm(theory_t th,int order,diagram_t *dm)
{
   int nv,nl,nt,ne;
   int i,j,iv1,iv2,ord,ie;
   int (*v)[2];
   line_t *l;
   vertex_t *vx;
   vlist_t vl;

   nv=(*dm).nv;
   nl=(*dm).nl;
   nt=(int)(LTYPES);

   if ((nv>nvs)||(nl>nls))
      alloc_aux(nv,nl);

   v=(*dm).v;
   l=(*dm).l;
   vl=vertices(th,order);
   ie=((nv<1)||(nl<0));
   ne=0;
   ord=0;

   for (i=0;i<nv;i++)
      ie|=((v[i][0]>order)||(v[i][1]>=vl.nv[v[i][0]]));

   for (i=0;i<nl;i++)
   {
      j=(int)(l[i].type);
      iv1=l[i].iv[0];
      iv2=l[i].iv[1];

      ie|=((j<0)||(j>=nt));
      ie|=((iv1<0)||(iv1>=nv));
      ie|=((iv2<0)||(iv2>=nv));
   }

   if (ie==0)
   {
      for (i=0;i<nv;i++)
      {
         vx=vl.v[v[i][0]]+v[i][1];
         ord+=(*vx).order;

         for (j=0;j<nt;j++)
            nla[i][j]=(*vx).nl[j];
      }

      for (i=0;i<nl;i++)
      {
         j=(int)(l[i].type);
         iv1=l[i].iv[0];
         iv2=l[i].iv[1];

         nla[iv1][j]-=1;
         nla[iv2][(int)(rvt[j])]-=1;
      }

      for (i=0;i<nv;i++)
      {
         for (j=0;j<nt;j++)
         {
            ie|=(nla[i][j]<0);
            ne+=nla[i][j];
         }
      }
   }

   error_loc((ie!=0)||(ord>order),1,"check_dgm [dmprop.c]",
             "Unexpected diagram data");

   return ne;
}


int is_tree(diagram_t *dm)
{
   int nv,nl,i;
   int (*v)[2];

   nv=(*dm).nv;
   nl=(*dm).nl;

   if (nl==(nv-1))
   {
      v=(*dm).v;

      for (i=0;i<nv;i++)
      {
         if (v[i][0]==0)
            return 0;
      }

      if (ncparts(dm)==1)
         return 1;
      else
         return 0;
   }
   else
      return 0;
}


int is_branch(diagram_t *dm)
{
   int nv,nl,i;
   int iv1,iv2;
   line_t *l;

   if (is_tree(dm)==0)
      return 0;

   nv=(*dm).nv;
   nl=(*dm).nl;
   l=(*dm).l;

   for (i=0;i<nv;i++)
      ivs[i]=0;

   for (i=0;i<nl;i++)
   {
      iv1=l[i].iv[0];
      iv2=l[i].iv[1];

      ivs[iv1]+=1;
      ivs[iv2]+=1;
   }

   for (i=0;i<nv;i++)
   {
      if (ivs[i]>2)
         return 0;
   }

   return 1;
}


static int next_part(int nv,int nl,line_t *l)
{
   int i,j,s;
   int iv1,iv2;

   for (i=0;i<nv;i++)
   {
      if (ivs[i]==0)
      {
         ivs[i]=1;
         break;
      }
   }

   if (i<nv)
   {
      s=1;

      while (s==1)
      {
         s=0;

         for (i=0;i<nl;i++)
         {
            if (ils[i]==0)
            {
               iv1=l[i].iv[0];
               iv2=l[i].iv[1];
               j=ivs[iv1]+ivs[iv2];

               if (j==1)
               {
                  ivs[iv1]=1;
                  ivs[iv2]=1;
                  s=1;
               }

               if (j>0)
                  ils[i]=1;
            }
         }
      }

      return 1;
   }

   return 0;
}


int ncparts(diagram_t *dm)
{
   int nv,nl,n,i;
   line_t *l;

   nv=(*dm).nv;
   nl=(*dm).nl;
   l=(*dm).l;
   n=0;

   if ((nv>nvs)||(nl>nls))
      alloc_aux(nv,nl);

   for (i=0;i<nv;i++)
      ivs[i]=0;

   for (i=0;i<nl;i++)
      ils[i]=0;

   while (next_part(nv,nl,l))
      n+=1;

   return n;
}


int nquark_loops(diagram_t *dm)
{
   int nv,nl,n,i;
   line_t *l;

   nv=(*dm).nv;
   nl=(*dm).nl;
   l=(*dm).l;
   n=0;

   if ((nv>nvs)||(nl>nls))
      alloc_aux(nv,nl);

   for (i=0;i<nv;i++)
      ivs[i]=0;

   for (i=0;i<nl;i++)
   {
      if ((l[i].type==QUARK)||(l[i].type==ANTIQUARK))
      {
         ils[i]=0;
         ivs[l[i].iv[0]]+=1;
         ivs[l[i].iv[1]]+=1;
      }
      else
         ils[i]=1;
   }

   for (i=0;i<nv;i++)
   {
      if (ivs[i]==0)
         ivs[i]=1;
      else
      {
         if (ivs[i]==1)
            n-=1;
         ivs[i]=0;
      }
   }

   n/=2;

   while (next_part(nv,nl,l))
      n+=1;

   return n;
}


int nghost_loops(diagram_t *dm)
{
   int nv,nl,n,i;
   line_t *l;

   nv=(*dm).nv;
   nl=(*dm).nl;
   l=(*dm).l;
   n=0;

   if ((nv>nvs)||(nl>nls))
      alloc_aux(nv,nl);

   for (i=0;i<nv;i++)
      ivs[i]=0;

   for (i=0;i<nl;i++)
   {
      if ((l[i].type==GHOST)||(l[i].type==ANTIGHOST))
      {
         ils[i]=0;
         ivs[l[i].iv[0]]+=1;
         ivs[l[i].iv[1]]+=1;
      }
      else
         ils[i]=1;
   }

   for (i=0;i<nv;i++)
   {
      if (ivs[i]==0)
         ivs[i]=1;
      else
      {
         if (ivs[i]==1)
            n-=1;
         ivs[i]=0;
      }
   }

   n/=2;

   while (next_part(nv,nl,l))
      n+=1;

   return n;
}


int nlext(diagram_t *dm,vlist_t *vl,int *n1pt)
{
   int nt,nv;
   int (*v)[2];
   int ne,n1,i,j,*nlv;
   vertex_t *vx;

   nt=(int)(LTYPES);
   nv=(*dm).nv;
   v=(*dm).v;
   ne=0;
   n1=0;

   for (i=0;i<nv;i++)
   {
      vx=(*vl).v[v[i][0]]+v[i][1];
      nlv=(*vx).nl;

      for (j=0;j<nt;j++)
         ne+=nlv[j];

      n1+=(v[i][0]==0);
   }

   (*n1pt)=n1;

   return ne-2*(*dm).nl;
}


int has_tadpole(diagram_t *dm,ltype_t lt,vlist_t *vl)
{
   int nv,nl,i;
   int ne,ne0,n1pt;
   line_t *l;

   nv=(*dm).nv;
   nl=(*dm).nl;

   if ((nv>nvs)||(nl>nls))
      alloc_aux(nv,nl);

   ne0=nlext(dm,vl,&n1pt)+2;
   l=(*dm).l;

   for (i=0;i<nl;i++)
   {
      if ((l[i].type==lt)&&(l[i].iv[0]!=l[i].iv[1]))
      {
         dmw[0][0].nv=nv;
         dmw[0][0].nl=nl;
         remove_line(i,dm,dmw[0]);

         dmw[1][0].nv=nv;
         dmw[1][0].nl=nl;
         connected_part(l[i].iv[0],dmw[0],dmw[1]);
         ne=nlext(dmw[1],vl,&n1pt);

         if ((ne==1)&&(n1pt==0))
            return 1;

         if (ne!=ne0)
         {
            dmw[1][0].nv=nv;
            dmw[1][0].nl=nl;
            connected_part(l[i].iv[1],dmw[0],dmw[1]);
            ne=nlext(dmw[1],vl,&n1pt);

            if ((ne==1)&&(n1pt==0))
               return 1;
         }
      }
   }

   return 0;
}


int nVX(diagram_t *dm,vtype_t VX,vlist_t *vl)
{
   int nv,n,i;
   int (*v)[2];
   vertex_t *vx;

   nv=(*dm).nv;
   v=(*dm).v;
   n=0;

   for (i=0;i<nv;i++)
   {
      vx=(*vl).v[v[0][0]]+v[0][1];
      n+=((*vx).type==VX);
      v+=1;
   }

   return n;
}
