
/*******************************************************************************
*
* File order.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Ordering of the diagram elements and diagram comparison.
*
* The externally accessible functions are
*
*   int *alloc_rank(int nv)
*     Allocates an array of integers of length equal to nv and returns
*     its address. The elements of the array are set to 0. The program
*     returns NULL if nv<=0.
*
*   void reorder_dgm(diagram_t *dm,int *rk)
*     Reorders the elements of the diagram dm according to the rules
*     described in the notes. The ranks of the reordered vertices are
*     assigned to the array elements rk[0],..,rk[nv-1], where nv is the
*     number of vertices in the diagram (the array is assumed to have
*     at least this many elements).
*
*   int rank_dgm(diagram_t *dm1,int *rk1,diagram_t *dm2,int *rk2)
*     Ranks the diagrams dm1 and dm2 based on a quick comparison of the
*     diagram data, assuming the diagram elements have been ordered by
*     the program reorder_dgm() and that rk1 and rk2 are the associated
*     arrays of vertex ranks. The program returns 1 if dm1<dm2 and -1
*     if dm1>dm2. In all other cases, the vertex, rank and line-type
*     arrays of the diagrams coincide and 0 is returned.
*
*   int cmp_dgm(diagram_t *dm1,int *rk1,diagram_t *dm2,int *rk2)
*     Compares the diagrams dm1 and dm2, assuming their elements have
*     been ordered by the program reorder_dgm() and that rk1 and rk2 are
*     the associated arrays of vertex ranks. The program returns 1 if
*     dm1 and dm2 coincide up to a possible reordering of the vertices
*     and lines. Otherwise 0 is returned.
*
* Notes:
*
* The structures of type diagram_t are described at the top of the file
* diagram.c. The program reorder_dgm() ranks the elements of a diagram in
* a way that does not depend on their initial labeling. It then reorders
* the lines and vertices so that the elements with lower rank come first.
* Elements with equal rank are ordered arbitrarily.
*
* The ranking of the lines proceeds by first converting all anti-fermion
* lines to fermion lines and simultaneously interchanging their endpoints.
* After that the lines are ranked according to their position in the list
* of line types.
*
* In the case of the vertices, the initial ranking is based on the type of
* the vertex and the numbers and types of the attached lines. An iterative
* process is then started, where the current ranks of the neighbouring
* vertices are used as ordering criterion for a finer ranking. The process
* comes to an end when the ranking did not change in the last iteration.
*
* The calculated ranks rk[0],..,rk[nv-1] of the vertices are such that
*
*   rk[0]=0,
*
*   rk[i]=Number of vertices with rank strictly less than rk[i].
*
* In general, the ranks are therefore not increasing in steps of 1.
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define ORDER_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "permutations.h"
#include "rvt.h"

static int nvs=0,nls=0,*ns,*is;
static int *ivs,*rks,*nns,*ils;
static int **rns,**rvs,(*vs)[2];
static line_t *ls;


int *alloc_rank(int nv)
{
   int i;
   int *rk;

   if (nv>0)
   {
      rk=malloc(nv*sizeof(*rk));
      error_loc(rk==NULL,1,"alloc_rank [order.c]",
                "Unable to allocate rank array");

      for (i=0;i<nv;i++)
         rk[i]=0;

      return rk;
   }
   else
      return NULL;
}


static void alloc_aux(int nv,int nl)
{
   int nt,n,i;

   if ((nvs>0)||(nls>0))
      free(ns);
   if (nvs>0)
   {
      free(rns);
      free(vs);
   }
   if (nls>0)
      free(ls);

   if (nv<nvs)
      nv=nvs;
   if (nl<nls)
      nl=nls;

   nt=(int)(LTYPES)+1;
   n=2*nt+3*nv+nv*nt+nv*nv+nl;
   ns=malloc(n*sizeof(*ns));
   rns=malloc(2*nv*sizeof(*rns));
   vs=malloc(nv*sizeof(*vs));

   if (nl>0)
      ls=malloc(nl*sizeof(*ls));

   error_loc((ns==NULL)||(rns==NULL)||(vs==NULL)||((nl>0)&&(ls==NULL)),1,
             "alloc_aux [order.c]","Unable to allocate auxiliary arrays");

   is=ns+nt;
   ivs=is+nt;
   rks=ivs+nv;
   nns=rks+nv;

   rvs=rns+nv;
   rns[0]=nns+nv;

   for (i=1;i<nv;i++)
      rns[i]=rns[i-1]+nt;

   rvs[0]=rns[0]+nv*nt;

   for (i=1;i<nv;i++)
      rvs[i]=rvs[i-1]+nv;

   if (nl>0)
      ils=rvs[0]+nv*nv;
   else
      ils=NULL;

   nvs=nv;
   nls=nl;

   for (i=0;i<n;i++)
      ns[i]=0;

   for (i=0;i<nv;i++)
   {
      vs[i][0]=0;
      vs[i][1]=0;
   }

   for (i=0;i<nl;i++)
   {
      ls[i].type=LTYPES;
      ls[i].iv[0]=0;
      ls[i].iv[1]=0;
   }
}


static void sort_int(int n,int *x)
{
   int i,j,s;

   for (i=0;i<(n-1);i++)
   {
      for (j=(i+1);j<n;j++)
      {
         if (x[i]>x[j])
         {
            s=x[i];
            x[i]=x[j];
            x[j]=s;
         }
      }
   }
}


static int cmp_int(int n,int *x,int *y)
{
   int i;

   for (i=0;i<n;i++)
   {
      if (x[i]<y[i])
         return 1;
      else if (x[i]>y[i])
         return -1;
   }

   return 0;
}


static void reorder_lines(diagram_t *dm)
{
   int nl,s,i,j;
   line_t *l;

   nl=(*dm).nl;
   l=(*dm).l;

   for (i=0;i<=(int)(LTYPES);i++)
      ns[i]=0;

   for (i=0;i<nl;i++)
   {
      if ((l[i].type==ANTIGHOST)||(l[i].type==ANTIQUARK))
      {
         l[i].type=rvt[(int)(l[i].type)];
         s=l[i].iv[0];
         l[i].iv[0]=l[i].iv[1];
         l[i].iv[1]=s;
      }

      ns[(int)(l[i].type)]+=1;
   }

   is[0]=0;

   for (i=1;i<=(int)(LTYPES);i++)
      is[i]=is[i-1]+ns[i-1];

   for (i=0;i<nl;i++)
   {
      j=is[(int)(l[i].type)];

      ls[j].type=l[i].type;
      ls[j].iv[0]=l[i].iv[0];
      ls[j].iv[1]=l[i].iv[1];

      is[(int)(l[i].type)]+=1;
   }

   for (i=0;i<nl;i++)
   {
      l[i].type=ls[i].type;
      l[i].iv[0]=ls[i].iv[0];
      l[i].iv[1]=ls[i].iv[1];
   }
}


static void cnt_lines(diagram_t *dm)
{
   int nv,nl,nt,i,j;
   line_t *l;

   nv=(*dm).nv;
   nl=(*dm).nl;
   l=(*dm).l;
   nt=(int)(LTYPES)+1;

   for (i=0;i<nv;i++)
   {
      for (j=0;j<nt;j++)
         rns[i][j]=0;
   }

   for (i=0;i<nl;i++)
   {
      rns[(*l).iv[0]][(int)((*l).type)]+=1;
      rns[(*l).iv[1]][(int)(rvt[(int)((*l).type)])]+=1;
      l+=1;
   }
}


static void nn_ranks(diagram_t *dm)
{
   int nv,nl,i,j;
   int i0,i1;
   line_t *l;

   nv=(*dm).nv;
   nl=(*dm).nl;
   l=(*dm).l;

   for (i=0;i<nv;i++)
   {
      for (j=0;j<nv;j++)
         rvs[i][j]=nv;
   }

   for (i=0;i<nl;i++)
   {
      i0=(*l).iv[0];
      i1=(*l).iv[1];

      rvs[i0][i1]=rks[i1];
      rvs[i1][i0]=rks[i0];

      l+=1;
   }

   for (i=0;i<nv;i++)
   {
      sort_int(nv,rvs[i]);

      for (j=0;j<nv;j++)
      {
         if (rvs[i][j]==nv)
            break;
      }

      nns[i]=j;
   }
}


static int cmp0(int i,int j)
{
   int (*v0)[2],(*v1)[2];

   v0=vs+i;
   v1=vs+j;

   if ((*v0)[0]<(*v1)[0])
      return 1;
   else if ((*v0)[0]>(*v1)[0])
      return -1;
   else if ((*v0)[1]<(*v1)[1])
      return 1;
   else if ((*v0)[1]>(*v1)[1])
      return -1;

   return 0;
}


static int cmp1(int i,int j)
{
   return cmp_int((int)(LTYPES),rns[i],rns[j]);
}


static int cmp2(int i,int j)
{
   int nn;

   nn=nns[i];

   if (nn<nns[j])
      nn=nns[j];

   return cmp_int(nn,rvs[i],rvs[j]);
}


static int rank_vertices(int nv,int (*cmp)(int i,int j))
{
   int i0,i1,i,j,k;
   int s,r;

   s=0;

   for (i0=0;i0<nv;)
   {
      r=0;
      i=rks[ivs[i0]];
      i1=i0+1;

      while ((i1<nv)&&(rks[ivs[i1]]==i))
         i1+=1;

      for (i=i0;i<(i1-1);i++)
      {
         for (j=(i+1);j<i1;j++)
         {
            k=cmp(ivs[i],ivs[j]);
            r|=(k!=0);

            if (k==1)
               rks[ivs[j]]+=1;
            else if (k==-1)
               rks[ivs[i]]+=1;
         }
      }

      if (r!=0)
      {
         for (i=i0;i<(i1-1);i++)
         {
            for (j=(i+1);j<i1;j++)
            {
               if (rks[ivs[i]]>rks[ivs[j]])
               {
                  k=ivs[i];
                  ivs[i]=ivs[j];
                  ivs[j]=k;
               }
            }
         }
      }

      i0=i1;
      s|=r;
   }

   return s;
}


static void reorder_vertices(diagram_t *dm,int *rk)
{
   int nv,nl,i,j,s;
   int (*v)[2];
   line_t *l;

   nv=(*dm).nv;
   nl=(*dm).nl;
   v=(*dm).v;
   l=(*dm).l;

   for (i=0;i<nv;i++)
   {
      ivs[i]=i;
      rks[i]=0;
      vs[i][0]=v[i][0];
      vs[i][1]=v[i][1];
   }

   (void)(rank_vertices(nv,cmp0));

   if (nl>0)
   {
      cnt_lines(dm);
      (void)(rank_vertices(nv,cmp1));
   }

   s=1;

   while (s==1)
   {
      nn_ranks(dm);
      s=rank_vertices(nv,cmp2);
   }

   for (i=0;i<nv;i++)
   {
      j=ivs[i];
      rk[i]=rks[j];
      v[i][0]=vs[j][0];
      v[i][1]=vs[j][1];
      rks[j]=i;
   }

   for (i=0;i<nl;i++)
   {
      l[i].iv[0]=rks[l[i].iv[0]];
      l[i].iv[1]=rks[l[i].iv[1]];
   }
}


void reorder_dgm(diagram_t *dm,int *rk)
{
   int nv,nl;

   nv=(*dm).nv;
   nl=(*dm).nl;

   if ((nv>nvs)||(nl>nls))
      alloc_aux(nv,nl);

   if (nl>0)
      reorder_lines(dm);

   if (nv>1)
      reorder_vertices(dm,rk);
   else if (nv==1)
      rk[0]=0;
}


int rank_dgm(diagram_t *dm1,int *rk1,diagram_t *dm2,int *rk2)
{
   int nv,nl,i;
   int (*v1)[2],(*v2)[2];
   line_t *l1,*l2;

   nv=(*dm1).nv;

   if (nv<(*dm2).nv)
      return 1;
   else if (nv>(*dm2).nv)
      return -1;

   nl=(*dm1).nl;

   if (nl<(*dm2).nl)
      return 1;
   else if (nl>(*dm2).nl)
      return -1;

   v1=(*dm1).v;
   v2=(*dm2).v;

   for (i=0;i<nv;i++)
   {
      if ((*v1)[0]<(*v2)[0])
         return 1;
      else if ((*v1)[0]>(*v2)[0])
         return -1;
      else if ((*v1)[1]<(*v2)[1])
         return 1;
      else if ((*v1)[1]>(*v2)[1])
         return -1;

      if ((*rk1)<(*rk2))
         return 1;
      else if ((*rk1)>(*rk2))
         return -1;

      v1+=1;
      v2+=1;
      rk1+=1;
      rk2+=1;
   }

   l1=(*dm1).l;
   l2=(*dm2).l;

   for (i=0;i<nl;i++)
   {
      if ((int)((*l1).type)<(int)((*l2).type))
         return 1;
      else if ((int)((*l1).type)>(int)((*l2).type))
         return -1;

      l1+=1;
      l2+=1;
   }

   return 0;
}


static int cmp_lines(int nl,line_t *l1,line_t *l2,int *p)
{
   int i,j,i0,i1;
   int iv10,iv11,iv20,iv21;
   ltype_t ab,ba;

   for (i=0;i<nl;i++)
      ils[i]=0;

   i0=0;
   i1=0;
   ab=(ltype_t)(0);
   ba=ab;

   for (i=0;i<nl;i++)
   {
      if (i==i1)
      {
         i0=i1;
         ab=l1[i0].type;
         ba=rvt[(int)(ab)];
         i1=i0+1;

         while ((i1<nl)&&(l1[i1].type==ab))
            i1+=1;
      }

      iv10=p[l1[i].iv[0]];
      iv11=p[l1[i].iv[1]];

      for (j=i0;j<i1;j++)
      {
         if (ils[j]==0)
         {
            iv20=l2[j].iv[0];
            iv21=l2[j].iv[1];

            if (((iv10==iv20)&&(iv11==iv21))||
                ((ba==ab)&&(iv10==iv21)&&(iv11==iv20)))
            {
               ils[j]=1;
               break;
            }
         }
      }

      if (j==i1)
         return 0;
   }

   return 1;
}


int cmp_dgm(diagram_t *dm1,int *rk1,diagram_t *dm2,int *rk2)
{
   int nv,nl,i;
   int m,*n,*p;
   long np,ip;

   if (rank_dgm(dm1,rk1,dm2,rk2)!=0)
      return 0;

   nv=(*dm1).nv;
   nl=(*dm1).nl;

   if (nl==0)
      return 1;

   if ((nv>nvs)||(nl>nls))
      alloc_aux(nv,nl);

   m=0;
   n=nns;
   n[0]=1;

   for (i=1;i<nv;i++)
   {
      if (rk1[i]==rk1[i-1])
         n[m]+=1;
      else
      {
         m+=1;
         n[m]=1;
      }
   }

   np=init_perm(m+1,n);
   ip=0L;
   p=rks;

   while (ip<np)
   {
      ip=next_perm(p);

      if (cmp_lines(nl,(*dm1).l,(*dm2).l,p))
      {
         stop_perm();
         return 1;
      }
   }

   stop_perm();
   return 0;
}
