
/*******************************************************************************
*
* File vertices.c
*
* Copyright (C) 2014, 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Vertices of the supported theories.
*
* The externally accessible function is
*
*   vlist_t vertices(theory_t th,int order)
*     Returns a structure containing the list of vertices in the
*     theory with label "th" up to and including the specified order
*     in the coupling (see the notes).
*
* Notes:
*
* Currently the supported theories and their name tags are:
*
*  PHI4                  Phi^4 theory.
*
*  YM                    Continuum pure gauge theory.
*
*  QCD                   Continuum QCD.
*
*  LYM                   Pure lattice gauge theory.
*
*  LQCD                  Lattice QCD.
*
* Vertices are described by structures with the elements
*
*  type                  Can be SOURCE,EXTERNAL,VS,VSM,VG,VGM,VGF,VFP,
*                        VQG or VQM where
*
*                          VS     scalar field vertex,
*                          VSM    scalar field mass insertion,
*                          VG     gluon vertex,
*                          VGM    gluon measure vertex (in LYM and LQCD),
*                          VGF    gauge-fixing vertex,
*                          VFP    Faddeev-Popov ghost vertex,
*                          VQG    quark-gluon vertex,
*                          VQM    quark mass insertion.
*
*  order                 Order in the coupling of the vertex.
*
*  nl[k]                 Number of lines of type k (k=0,..,(int)(LTYPES)-1).
*
* The possible line types are SCALAR,GLUON,GHOST,ANTIGHOST,QUARK and
* ANTIQUARK.
*
* The structure returned by vertices() has the following elements
*
*  order                 Maximal order in the coupling of the vertices
*                        included in the structure.
*
*  nv[k]                 Number of vertices at order k in the coupling
*                        (k=0,..,order).
*
*  v[k][j]               Array of vertices at order k (k=0,..,order,
*                        j=0,..,nv[k]-1).
*
* Note that the value of "order" in the structure may be smaller than the one
* passed through the argument list of the program if the theory has vertices
* only up to some finite order.
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define VERTICES_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "diagrams.h"

static vlist_t vls[THEORIES]={{-1,NULL,NULL},{-1,NULL,NULL},{-1,NULL,NULL},
                              {-1,NULL,NULL},{-1,NULL,NULL}};


static void set_vertices(theory_t th,int order)
{
   int i,j,k,n,*nv;
   vertex_t **vv,*v;
   vlist_t *vl;

   nv=malloc((order+1)*sizeof(*nv));
   vv=malloc((order+1)*sizeof(*vv));
   error_loc((nv==NULL)||(vv==NULL),1,"set_vertices [vertices.c]",
             "Unable to allocate vertex array");

   for (i=0;i<=order;i++)
   {
      nv[i]=0;
      vv[i]=NULL;
   }

   if (th==PHI4)
   {
      nv[0]=2;
      if (order>0)
         nv[1]=2;
   }
   else if (th==YM)
   {
      nv[0]=4;
      if (order>0)
         nv[1]=2;
      if (order>1)
         nv[2]=2;
   }
   else if (th==QCD)
   {
      nv[0]=6;
      if (order>0)
         nv[1]=3;
      if (order>1)
         nv[2]=3;
   }
   else if (th==LYM)
   {
      nv[0]=4;
      if (order>0)
         nv[1]=2;
      if (order>1)
         nv[2]=4;
   }
   else if (th==LQCD)
   {
      nv[0]=6;
      if (order>0)
         nv[1]=3;
      if (order>1)
         nv[2]=6;
   }
   else
      error_loc(1,1,"set_vertices [vertices.c]","Unknown type of theory");

   if (th==LYM)
   {
      for (i=3;i<=order;i++)
      {
         if ((i%2)==0)
            nv[i]=3;
         else
            nv[i]=1;
      }
   }
   else if (th==LQCD)
   {
      for (i=3;i<=order;i++)
      {
         if ((i%2)==0)
            nv[i]=4;
         else
            nv[i]=2;
      }
   }

   n=0;

   for (i=0;i<=order;i++)
      n+=nv[i];

   v=malloc(n*sizeof(*v));
   error_loc(v==NULL,1,"set_vertices [vertices.c]",
             "Unable to allocate vertex array");

   for (i=0;i<=order;i++)
   {
      if (nv[i]>0)
      {
         vv[i]=v;
         v+=nv[i];
      }
   }

   for (i=0;i<=order;i++)
   {
      for (j=0;j<nv[i];j++)
      {
         vv[i][j].order=i;

         for (k=0;k<(int)(LTYPES);k++)
            vv[i][j].nl[k]=0;
      }

      j=0;

      if (i==0)
      {
         if (th==PHI4)
         {
            vv[i][j].type=SOURCE;
            vv[i][j].nl[(int)(SCALAR)]=1;
            j+=1;

            vv[i][j].type=EXTERNAL;
            vv[i][j].nl[(int)(SCALAR)]=1;
         }
         else
         {
            vv[i][j].type=SOURCE;
            vv[i][j].nl[(int)(GLUON)]=1;
            j+=1;

            vv[i][j].type=EXTERNAL;
            vv[i][j].nl[(int)(GLUON)]=1;
            j+=1;

            vv[i][j].type=SOURCE;
            vv[i][j].nl[(int)(GHOST)]=1;
            j+=1;

            vv[i][j].type=SOURCE;
            vv[i][j].nl[(int)(ANTIGHOST)]=1;
            j+=1;

            if ((th==QCD)||(th==LQCD))
            {
               vv[i][j].type=SOURCE;
               vv[i][j].nl[(int)(QUARK)]=1;
               j+=1;

               vv[i][j].type=SOURCE;
               vv[i][j].nl[(int)(ANTIQUARK)]=1;
            }
         }
      }
      else if (i==1)
      {
         if (th==PHI4)
         {
            vv[i][j].type=VS;
            vv[i][j].nl[(int)(SCALAR)]=4;
            j+=1;

            vv[i][j].type=VSM;
            vv[i][j].nl[(int)(SCALAR)]=2;
         }
         else
         {
            vv[i][j].type=VG;
            vv[i][j].nl[(int)(GLUON)]=3;
            j+=1;

            vv[i][j].type=VFP;
            vv[i][j].nl[(int)(GLUON)]=1;
            vv[i][j].nl[(int)(GHOST)]=1;
            vv[i][j].nl[(int)(ANTIGHOST)]=1;
            j+=1;

            if ((th==QCD)||(th==LQCD))
            {
               vv[i][j].type=VQG;
               vv[i][j].nl[(int)(GLUON)]=1;
               vv[i][j].nl[(int)(QUARK)]=1;
               vv[i][j].nl[(int)(ANTIQUARK)]=1;
            }
         }
      }
      else if (i==2)
      {
         vv[i][j].type=VG;
         vv[i][j].nl[(int)(GLUON)]=4;
         j+=1;

         vv[i][j].type=VGF;
         vv[i][j].nl[(int)(GLUON)]=2;
         j+=1;

         if ((th==QCD)||(th==LQCD))
         {
            vv[i][j].type=VQM;
            vv[i][j].nl[(int)(QUARK)]=1;
            vv[i][j].nl[(int)(ANTIQUARK)]=1;
            j+=1;
         }

         if ((th==LYM)||(th==LQCD))
         {
            vv[i][j].type=VGM;
            vv[i][j].nl[(int)(GLUON)]=2;
            j+=1;

            vv[i][j].type=VFP;
            vv[i][j].nl[(int)(GLUON)]=2;
            vv[i][j].nl[(int)(GHOST)]=1;
            vv[i][j].nl[(int)(ANTIGHOST)]=1;
            j+=1;

            if (th==LQCD)
            {
               vv[i][j].type=VQG;
               vv[i][j].nl[(int)(GLUON)]=2;
               vv[i][j].nl[(int)(QUARK)]=1;
               vv[i][j].nl[(int)(ANTIQUARK)]=1;
            }
         }
      }
      else if ((th==LYM)||(th==LQCD))
      {
         vv[i][j].type=VG;
         vv[i][j].nl[(int)(GLUON)]=i+2;
         j+=1;

         if ((i%2)==0)
         {
            vv[i][j].type=VGM;
            vv[i][j].nl[(int)(GLUON)]=i;
            j+=1;

            vv[i][j].type=VFP;
            vv[i][j].nl[(int)(GLUON)]=i;
            vv[i][j].nl[(int)(GHOST)]=1;
            vv[i][j].nl[(int)(ANTIGHOST)]=1;
            j+=1;
         }

         if (th==LQCD)
         {
            vv[i][j].type=VQG;
            vv[i][j].nl[(int)(GLUON)]=i;
            vv[i][j].nl[(int)(QUARK)]=1;
            vv[i][j].nl[(int)(ANTIQUARK)]=1;
         }
      }
   }

   vl=vls+(int)(th);

   if ((*vl).nv!=NULL)
   {
      free((*vl).nv);
      free((*vl).v[0]);
      free((*vl).v);
   }

   (*vl).order=order;
   (*vl).nv=nv;
   (*vl).v=vv;
}


vlist_t vertices(theory_t th,int order)
{
   vlist_t vl;

   error_loc((th==THEORIES)||(order<0),1,"vertices [vertices.c]",
             "Arguments are out of range");

   if ((th==PHI4)&&(order>1))
      order=1;
   else if (((th==YM)||(th==QCD))&&(order>2))
      order=2;

   if (vls[(int)(th)].order<order)
      set_vertices(th,order);

   vl=vls[(int)(th)];
   vl.order=order;

   return vl;
}
