
/*******************************************************************************
*
* File kfact.c
*
* Copyright (C) 2014, 2016 Martin Luescher
*
* Normalization of the gradient-flow running coupling.
*
* The externally accessible function is
*
*   double kfact(char *type,int N,int T,int L,int bc,double c1,
*                int x0,double t)
*     Returns the normalization factor k of the gradient-flow coupling
*     (see the notes).
*
* Notes:
*
* For Schroedinger functional (SF) and open-SF boundary conditions, and all
* gauge actions supported in openQCD-1.4, the normalization factor of the
* gradient-flow coupling is computed in the notes
*
*  M. Luescher:
*  "Normalization of the gradient-flow coupling in LQCD"
*
* (file doc/gfnorm.pdf).
*
* The parameters of kfact() are
*
*  type              Specifies the type of coupling ("c","p","cs","ct",
*                    "ps","pt", where "c" stands for clover, "p" for
*                    plaquette, "s" for spatial and "t" for time-like).
*
*  N                 Number of colours.
*
*  T,L               Time-like and spatial extent of the lattice.
*
*  bc                Boundary condition (1: SF, 2: open-SF).
*
*  c1                Coefficient of the double-plaquette terms in the
*                    gauge action.
*
*  x0                Time at which the gradient-flow coupling is
*                    defined (0<x0<T).
*
*  t                 Flow time.
*
* All dimensionful quantities must be given in lattice units. The coefficient
* c in the notes is related to the flow time t through
*
*  c=sqrt(8*t)/L
*
* In an MPI program, this function should only be called by the MPI root
* process.
*
*******************************************************************************/

#define KFACT_C

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "utils.h"
#include "gbar.h"

static int Ts=0,Ls=0;
static double c1s,vm[4],rm[4],Dmn[4][4],Smn[4][4];
static double *ph0,*ph1,*sn,*cn;


static void alloc_arrays(int T,int L)
{
   if (Ts>0)
      free(ph0);

   ph0=malloc((3*T+L)*sizeof(*ph0));
   error_root(ph0==NULL,1,"alloc_arrays [kfact.c]",
              "Unable to allocate auxiliary arrays");

   ph1=ph0+T;
   sn=ph1+L;
   cn=sn+T;

   Ts=T;
   Ls=L;
}


static void set_phsq(char *type,int T,int L,int bc,int x0)
{
   int n;
   double pi,p0,x0s,r;

   pi=4.0*atan(1.0);
   x0s=(double)(x0);

   for (n=0;n<T;n++)
   {
      if (bc==1)
         p0=(double)(n)*pi/(double)(T);
      else
         p0=((double)(n)+0.5)*pi/(double)(T);

      ph0[n]=2.0*sin(0.5*p0);

      if (type[0]=='c')
      {
         r=sin(p0*x0s);
         sn[n]=r*r;
         r=cos(p0*x0s);
         cn[n]=r*r;
      }
      else if (bc==1)
      {
         r=sin(p0*x0s);
         sn[n]=r*r;
         r=cos(p0*(x0s+0.5));
         cn[n]=r*r;
         r=cos(p0*(x0s-0.5));
         cn[n]+=r*r;
         cn[n]*=0.5;
      }
      else
      {
         r=sin(p0*(x0s+0.5));
         sn[n]=r*r;
         r=sin(p0*(x0s-0.5));
         sn[n]+=r*r;
         sn[n]*=0.5;
         r=cos(p0*x0s);
         cn[n]=r*r;
      }
   }

   for (n=0;n<L;n++)
      ph1[n]=2.0*sin((double)(n)*pi/(double)(L));
}


static double set_Smn(char *type,int T,int bc,int *n)
{
   int mu,nu;
   double ph[4],phs[4];
   double ph2,ph4,sm;

   ph[0]=ph0[n[0]];
   ph[1]=ph1[n[1]];
   ph[2]=ph1[n[2]];
   ph[3]=ph1[n[3]];

   phs[0]=ph[0]*ph[0];
   phs[1]=ph[1]*ph[1];
   phs[2]=ph[2]*ph[2];
   phs[3]=ph[3]*ph[3];

   ph2=phs[0]+phs[1]+phs[2]+phs[3];
   ph4=phs[0]*phs[0]+phs[1]*phs[1]+phs[2]*phs[2]+phs[3]*phs[3];

   rm[0]=ph2-c1s*(ph4+ph2*phs[0]);
   rm[1]=ph2-c1s*(ph4+ph2*phs[1]);
   rm[2]=ph2-c1s*(ph4+ph2*phs[2]);
   rm[3]=ph2-c1s*(ph4+ph2*phs[3]);

   vm[0]=c1s*ph[0]*phs[0]/rm[0];
   vm[1]=c1s*ph[1]*phs[1]/rm[1];
   vm[2]=c1s*ph[2]*phs[2]/rm[2];
   vm[3]=c1s*ph[3]*phs[3]/rm[3];

   sm=vm[0]*vm[0]*rm[0]+vm[1]*vm[1]*rm[1]+
      vm[2]*vm[2]*rm[2]+vm[3]*vm[3]*rm[3];
   sm=1.0/(1.0+sm);

   if ((n[0]!=0)||(bc==2))
   {
      Dmn[0][0]=1.0/rm[0]-sm*vm[0]*vm[0];
      Dmn[1][1]=1.0/rm[1]-sm*vm[1]*vm[1];
      Dmn[2][2]=1.0/rm[2]-sm*vm[2]*vm[2];
      Dmn[3][3]=1.0/rm[3]-sm*vm[3]*vm[3];

      Dmn[0][1]=-sm*vm[0]*vm[1];
      Dmn[1][0]=Dmn[0][1];

      Dmn[0][2]=-sm*vm[0]*vm[2];
      Dmn[2][0]=Dmn[0][2];

      Dmn[0][3]=-sm*vm[0]*vm[3];
      Dmn[3][0]=Dmn[0][3];

      Dmn[1][2]=-sm*vm[1]*vm[2];
      Dmn[2][1]=Dmn[1][2];

      Dmn[1][3]=-sm*vm[1]*vm[3];
      Dmn[3][1]=Dmn[1][3];

      Dmn[2][3]=-sm*vm[2]*vm[3];
      Dmn[3][2]=Dmn[2][3];
   }
   else
   {
      if ((n[1]!=0)||(n[2]!=0)||(n[3]!=0))
         Dmn[0][0]=1.0/(ph2-c1s*ph4);
      else
         Dmn[0][0]=(double)(T*T);

      Dmn[0][1]=0.0;
      Dmn[0][2]=0.0;
      Dmn[0][3]=0.0;

      Dmn[1][0]=0.0;
      Dmn[1][1]=0.0;
      Dmn[1][2]=0.0;
      Dmn[1][3]=0.0;

      Dmn[2][0]=0.0;
      Dmn[2][1]=0.0;
      Dmn[2][2]=0.0;
      Dmn[2][3]=0.0;

      Dmn[3][0]=0.0;
      Dmn[3][1]=0.0;
      Dmn[3][2]=0.0;
      Dmn[3][3]=0.0;
   }

   for (mu=0;mu<4;mu++)
   {
      Smn[mu][mu]=0.0;

      for (nu=(mu+1);nu<4;nu++)
      {
         Smn[mu][nu]=
            phs[mu]*Dmn[nu][nu]+phs[nu]*Dmn[mu][mu]
            -2.0*ph[mu]*ph[nu]*Dmn[mu][nu];

         if (type[0]=='c')
            Smn[mu][nu]*=(1.0-0.25*phs[mu])*(1.0-0.25*phs[nu]);

         Smn[nu][mu]=Smn[mu][nu];
      }
   }

   return ph2;
}


double kfact(char *type,int N,int T,int L,int bc,double c1,int x0,double t)
{
   int nc,n[4];
   double sm,ph2,r,rs,rt;

   nc=strlen(type);

   error_root((nc<1)||(nc>2)||((type[0]!='c')&&(type[0]!='p'))||
              ((nc>1)&&(type[1]!='s')&&(type[1]!='t')),1,"kfact [kfact.c]",
              "Unknown type of running coupling");
   error_root((N<2)||(T<4)||(L<4),1,"kfact [kfact.c]",
              "N,T or L is out of range");
   error_root((bc<1)||(bc>2),1,"kfact [kfact.c]",
              "bc must be 1 or 2");
   error_root((8.0*c1)>=1.0,1,"kfact [kfact.c]",
              "c1 is out of range");
   error_root((x0<=0)||(x0>=T),1,"kfact [kfact.c]",
              "x0 is out of range");
   error_root(t<=0.0,1,"kfact [kfact.c]",
              "t must be positive");

   c1s=c1;
   if ((T>Ts)||(L>Ls))
      alloc_arrays(T,L);
   set_phsq(type,T,L,bc,x0);
   sm=0.0;

   for (n[0]=0;n[0]<T;n[0]++)
   {
      for (n[1]=0;n[1]<L;n[1]++)
      {
         for (n[2]=0;n[2]<L;n[2]++)
         {
            for (n[3]=0;n[3]<L;n[3]++)
            {
               ph2=set_Smn(type,T,bc,n);
               rs=Smn[2][1]+Smn[3][1]+Smn[3][2];
               rt=Smn[1][0]+Smn[2][0]+Smn[3][0];

               if (bc==1)
               {
                  if (n[0]!=0)
                  {
                     rs*=sn[n[0]];
                     rt*=cn[n[0]];
                  }
                  else
                  {
                     rs*=0.5*sn[n[0]];
                     rt*=0.5*cn[n[0]];
                  }
               }
               else
               {
                  rs*=cn[n[0]];
                  rt*=sn[n[0]];
               }

               if (nc==1)
                  r=rs+rt;
               else if (type[1]=='s')
                  r=rs;
               else
                  r=rt;

               sm+=r*exp(-2.0*t*ph2);
            }
         }
      }
   }

   sm*=t*t*(double)(N*N-1);
   sm/=((double)(T*L)*(double)(L*L));

   return 1.0/sm;
}
