
/*******************************************************************************
*
* File partition.c
*
* Copyright (C) 2014 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generation of partitions of an integer.
*
*   long init_partition(int n,int s)
*     Initializes the partition generator and returns the number of
*     partitions of s in n natural numbers (see the notes).
*
*   long next_partition(int *i)
*     Assigns the next partition to the array i and returns the current
*     partition count (see the notes). The partition counting starts with
*     1 and the generator wraps around when all possible partitions have
*     been generated.
*
*   void stop_partition(void)
*     Resets the partition generator to its uninitialized state.
*
* The partition generator runs through all vectors
*
*   i[0],i[1],..,i[n-1]
*
* of n integers satisfying
*
*   i[0]>=i[1]>=..>=i[n-1]>=1,
*
*   i[0]+i[1]+..+i[n-1]=s.
*
* The arguments n,s of the initialization program init_partition() must
* satisfy the constraints
*
*  n>0 and s>=n
*
* and be such that the number of partitions is less than or equal to
* LONG_MAX (on x86-64 machines, this is 2^63-1 and thus about 9.2x10^18).
* The program init_partition() aborts with an error message if these
* conditions are not satisfied.
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define PARTITION_C

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "utils.h"
#include "permutations.h"

static int na=0,ns,ss,init=0;
static int *is,*smi;
static long ni,nim;


static void set_arrays(int n,int s)
{
   int k;

   if (n>na)
   {
      if (na>0)
         free(is);
      na=n;
      is=malloc(2*na*sizeof(*is));
   }

   smi=is+n;
   ns=n;
   ss=s;

   is[0]=s;
   smi[0]=0;

   for (k=1;k<n;k++)
   {
      is[k]=0;
      smi[k]=0;
   }
}


static long npart(int n,int s)
{
   int i,imx;
   long sm,np;

   if ((n==1)||(s==0))
      return 1L;

   imx=s/n;
   sm=0L;

   for (i=0;i<=imx;i++)
   {
      np=npart(n-1,s-i*n);

      if (sm<=(LONG_MAX-np))
         sm+=np;
      else
         error_loc(1,1,"npart [partition.c]",
                   "Number of partitions is out of range");
   }

   return sm;
}


static void assign_idx(int *i)
{
   int k;

   for (k=0;k<ns;k++)
      i[k]=is[k]+1;
}


static int cycle(int k)
{
   int l;

   if ((smi[k]+(k+1)*(is[k]+1))<=ss)
   {
      is[k]+=1;

      for (l=(k-1);l>0;l--)
      {
         is[l]=is[k];
         smi[l]=smi[l+1]+is[k];
      }

      smi[0]=is[k]+smi[1];
      is[0]=ss-smi[0];

      return 1;
   }
   else if (k<(ns-1))
      return cycle(k+1);
   else
   {
      is[0]=ss;
      smi[0]=0;

      for (l=1;l<ns;l++)
      {
         is[l]=0;
         smi[l]=0;
      }

      return 0;
   }
}


long init_partition(int n,int s)
{
   error_loc(init==1,1,"init_partition [partition.c]",
             "A running partition generator cannot be re-initialized");

   error_loc((n<1)||(s<n),1,"init_partition [partition.c]",
             "Arguments are out of range");

   s-=n;
   set_arrays(n,s);
   nim=npart(n,s);

   ni=0L;
   init=1;

   return nim;
}


long next_partition(int *i)
{
   int ic;

   if (init!=1)
      error_loc(1,1,"next_partition [partition.c]",
                "Partition generator is not initialized");

   assign_idx(i);

   if (ns>1)
   {
      ic=cycle(1);
      ni+=1L;
   }
   else
      return 1;

   if (ic==1)
      return ni;
   else
   {
      ni=0L;
      return nim;
   }
}


void stop_partition(void)
{
   if (init==1)
      init=0;
   else
      error_loc(1,1,"stop_partition [partition.c]",
                "Partition generator is not initialized or already stopped");
}
