
/*******************************************************************************
*
* File rtree.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Recursive representation of rooted labeled tree diagrams.
*
*   rtree_t *get_rtree(theory_t th,int order,ltree_t *td)
*     Returns the address of an rtree_t structure that describes the
*     rooted labeled tree diagram "td" in the theory "th" at or below
*     the specified order in the coupling.
*
* Rooted labeled tree diagrams may be described recursively through a linked
* list of rtree_t structures. There is one structure per vertex and they are
* linked through the lines of the diagram. The representation is recursive,
* since each vertex can be considered to be at the root of a subtree, whose
* root line is the line that goes in direction of the root of the diagram.
*
* The elements of an rtree_t structure are:
*
*   nsrc        Number of source fields attached to the vertex.
*
*   nbr         Number of branches attached to the vertex. The
*               branch going to the root of the diagram is omitted.
*
*   v[2]        Indices of the vertex in the vertex list of the
*               theory considered [diagrams/vertices.c].
*
*   lt          Type of the root line at the vertex. By definition,
*               the root line is outgoing from the vertex.
*
*   src         Array of src_t structures describing the attached
*               source fields (nsrc elements).
*
*   br          Array of rtree_t structures describing the attached
*               branches (nbr elements).
*
* The elements of a src_t structure are:
*
*   lt          Source field type = type of the line from the source
*               to the vertex.
*
*   isrc        Label of the source field.
*
* By following the links to the branches, the vertices of a tree diagram are
* visited recursively. The program get_rtree() returns the rtree_t structure
* at the root of the tree diagram.
*
* The program in this module does not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define RTREE_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "rvt.h"
#include "trees.h"

static int na,nv,nl,(*v)[2];
static int nsrc,nbr;
static int ns[(int)(LTYPES)];
static line_t *la,*l;
static vlist_t vls;

static void next_rtree(line_t *rl,rtree_t *rt);


static rtree_t *alloc_rtree(int n)
{
   int i;
   rtree_t *rt;

   rt=malloc(n*sizeof(*rt));

   error_loc(rt==NULL,1,"alloc_rtree [rtree.c]",
             "Unable to allocate rtree_t structures");

   for (i=0;i<n;i++)
   {
      rt[i].nsrc=0;
      rt[i].nbr=0;
      rt[i].v[0]=0;
      rt[i].v[1]=0;
      rt[i].lt=LTYPES;
      rt[i].src=NULL;
      rt[i].br=NULL;
   }

   return rt;
}


static void set_nlegs(line_t *rl)
{
   int iv,il,it,ir;
   vertex_t *vx;

   nsrc=0;
   nbr=0;
   iv=(*rl).iv[0];
   vx=vls.v[v[iv][0]]+v[iv][1];

   for (it=0;it<(int)(LTYPES);it++)
      ns[it]=(*vx).nl[it];

   for (il=0;il<nl;il++)
   {
      if (l[il].iv[0]==iv)
      {
         if (l[il].iv[1]!=(*rl).iv[1])
            nbr+=1;
         it=(int)(l[il].type);
         ns[it]-=1;
      }

      if (l[il].iv[1]==iv)
      {
         if (l[il].iv[0]!=(*rl).iv[1])
            nbr+=1;
         it=(int)(l[il].type);
         ir=(int)(rvt[it]);
         ns[ir]-=1;
      }
   }

   if ((*rl).iv[1]==nv)
   {
      it=(int)((*rl).type);
      ns[it]-=1;
   }

   for (it=0;it<(int)(LTYPES);it++)
      nsrc+=ns[it];
}


static void set_src(line_t *rl,src_t *src)
{
   int iv,il,jl,it,ir,isrc;

   iv=(*rl).iv[0];

   for (il=0;il<na;il++)
   {
      if (la[il].iv[0]==iv)
      {
         it=(int)(la[il].type);
         ir=(int)(rvt[it]);
         ns[it]-=1;
         isrc=1;

         for (jl=0;jl<il;jl++)
         {
            if (((int)(la[jl].type)==it)||((int)(la[jl].type)==ir))
               isrc+=1;
         }

         (*src).lt=(ltype_t)(ir);
         (*src).isrc=isrc;
         src+=1;
      }

      if (la[il].iv[1]==iv)
      {
         it=(int)(la[il].type);
         ir=(int)(rvt[it]);
         ns[ir]-=1;
         isrc=1;

         for (jl=0;jl<il;jl++)
         {
            if (((int)(la[jl].type)==it)||((int)(la[jl].type)==ir))
               isrc+=1;
         }

         (*src).lt=(ltype_t)(it);
         (*src).isrc=isrc;
         src+=1;
      }
   }

   for (it=0;it<(int)(LTYPES);it++)
   {
      for (il=0;il<ns[it];il++)
      {
         (*src).lt=rvt[it];
         (*src).isrc=0;
         src+=1;
      }
   }
}


static void set_br(line_t *rl,rtree_t *br)
{
   int iv,il,it;
   line_t rlb;

   iv=(*rl).iv[0];
   rlb.iv[1]=iv;

   for (il=0;il<nl;il++)
   {
      if ((l[il].iv[0]==iv)&&(l[il].iv[1]!=(*rl).iv[1]))
      {
         rlb.iv[0]=l[il].iv[1];
         it=(int)(l[il].type);
         rlb.type=rvt[it];
         next_rtree(&rlb,br);
         br+=1;
      }

      if ((l[il].iv[1]==iv)&&(l[il].iv[0]!=(*rl).iv[1]))
      {
         rlb.iv[0]=l[il].iv[0];
         rlb.type=l[il].type;
         next_rtree(&rlb,br);
         br+=1;
      }
   }
}


static void next_rtree(line_t *rl,rtree_t *rt)
{
   int iv;
   src_t *src;
   rtree_t *br;

   iv=(*rl).iv[0];
   (*rt).v[0]=v[iv][0];
   (*rt).v[1]=v[iv][1];
   (*rt).lt=(*rl).type;

   set_nlegs(rl);

   if (nsrc>0)
      src=malloc(nsrc*sizeof(*src));
   else
      src=NULL;

   if (nbr>0)
      br=alloc_rtree(nbr);
   else
      br=NULL;

   error_loc(((nsrc>0)&&(src==NULL))||((nbr>0)&&(br==NULL)),1,
      "next_rtree [rtree.c]","Unable to allocate source or branch array");

   (*rt).nsrc=nsrc;
   (*rt).nbr=nbr;
   (*rt).src=src;
   (*rt).br=br;

   if (nsrc>0)
      set_src(rl,src);

   if (nbr>0)
      set_br(rl,br);
}


rtree_t *get_rtree(theory_t th,int order,ltree_t *td)
{
   int rv;
   line_t rl;
   diagram_t *dm;
   rtree_t *rt;

   vls=vertices(th,order);

   rv=(*td).rv;
   na=(*td).nl;
   la=(*td).l;
   dm=(*td).dm;

   nv=(*dm).nv;
   nl=(*dm).nl;
   v=(*dm).v;
   l=(*dm).l;

   error_loc(rv==nv,1,"get_rtree [rtree.c]",
             "Tree diagram is not rooted");

   rl.iv[0]=rv;
   rl.iv[1]=nv;

   if (th==PHI4)
      rl.type=SCALAR;
   else
      rl.type=GLUON;

   rt=alloc_rtree(1);
   next_rtree(&rl,rt);

   return rt;
}
