
/*******************************************************************************
*
* File rtree_gen.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Conversion of rooted labeled tree diagrams to the recursive representation.
*
*   int rtrees(theory_t th,int order)
*     Runs through all rooted labeled tree diagrams in the theory "th" up
*     to the given order in the coupling and converts the diagrams with
*     non-zero coefficients to the recursive representation described in
*     the file rtree.c. The converted diagrams are stored in static arrays
*     that can be accessed using other functions in this module (see the
*     notes). The program returns the total number of converted diagrams.
*
*   rcls_t *rtree_classes(theory_t th)
*     Returns a structure describing the classes of diagrams generated and
*     stored by the program rtrees() for the theory "th" (see the notes).
*     An error occurs if rtrees() has not previously been called.
*
*   rlist_t *select_rtrees(theory_t th,int icl)
*     Returns the pointer to an rlist_t structure containing the diagrams
*     in the class with index "icl" generated and stored by rtrees() for
*     the theory "th". The index "icl" is the one of the class in the list
*     returned by rtree_classes() (see the notes).
*
* The type rlist_t is described at the top of the file dmlist/rdlist.c. For
* a given theory, the list of tree diagrams constructed by rtrees() cannot
* be modified by calling rtrees() a second time with a different value of
* the parameter "order". An error occurs if the specified order is less
* than or equal to zero.
*
* The program rtrees() assumes that the list of rooted labeled tree diagrams
* has already been generated and that the root lines are set. An error occurs
* if this is not the case. Note, however, that rtrees() discards all diagrams
* with vanishing coefficient. The coefficients should therefore be computed
* (through tmap() [contract/tmap.c], for example) before rtrees() is called.
*
* The diagrams stored by rtrees() are divided into classes characterized by
* tupels
*
*   [order,n[0],..,n[nvt-1]]
*
* of non-negative integers, where
*
*   order         Order in the coupling of the diagrams in the class.
*
*   n[0],n[1],..  Numbers of "counterterm vertices".
*
*   nvt           Maximal number of types of "counterterm vertices" the
*                 diagrams stored by rtrees() can have.
*
* Depending on the theory, only a subset of all "counterterm vertices"
*
*   VSM,VGF,VQM,...
*
* can occur and the number nvt+1 of elements of the tupels characterizing the
* diagram classes thus vary. Exactly which vertices are included in the subset
* can be determined by calling rtree_classes(). The elements of the structure
* returned by this program are
*
*   nvt           Maximal number of types of "counterterm vertices".
*
*   vt            Array vt[0],..,vt[nvt-1] of the types of the "counterterm
*                 vertices".
*
*   cls           Address of a tplist_t structure containing the tupels
*                 characterizing the available classes of diagrams. The
*                 tupels are ordered in ascending order (see tplist.c
*                 in nompi/tupels).
*
* The number of vertices of type vt[i] in the diagrams contained in the class
* characterized by the tupel [order,n[0],..,n[nvt-1]] is equal to n[i].
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define RTREE_GEN_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "tupel.h"
#include "trees.h"

typedef struct
{
   int order;
   rcls_t *rc;
   rlist_t **rl;
} tsets_t;

static tsets_t trs[(int)(THEORIES)]={{0,NULL,NULL},
                                     {0,NULL,NULL},
                                     {0,NULL,NULL},
                                     {0,NULL,NULL},
                                     {0,NULL,NULL}};


static vtype_t *alloc_vt(int nvt)
{
   vtype_t *vt;

   vt=malloc(nvt*sizeof(*vt));
   error_loc(vt==NULL,1,"alloc_vt [rtree_gen.c]",
             "Unable to allocate vertex type array");
   return vt;
}


static void set_vt(theory_t th,int order)
{
   int nvt;
   vtype_t *vt;
   rcls_t *rc;

   nvt=0;
   vt=NULL;

   if (th==PHI4)
   {
      nvt=1;
      vt=alloc_vt(nvt);
      vt[0]=VSM;
   }
   else if ((th==YM)||(th==LYM))
   {
      if (order>=2)
      {
         nvt=1;
         vt=alloc_vt(nvt);
         vt[0]=VGF;
      }
   }
   else if ((th==QCD)||(th==LQCD))
   {
      if (order>=4)
      {
         nvt=2;
         vt=alloc_vt(nvt);
         vt[0]=VGF;
         vt[1]=VQM;
      }
      else if (order>=2)
      {
         nvt=1;
         vt=alloc_vt(nvt);
         vt[0]=VGF;
      }
   }

   rc=malloc(1*sizeof(*rc));
   error_loc(rc==NULL,1,"set_vt [rtree_gen.c]",
             "Unable to allocate rtree classes structure");

   (*rc).nvt=nvt;
   (*rc).vt=vt;
   trs[(int)(th)].rc=rc;
}


static void set_cls(theory_t th,int order)
{
   int i,j,k,iv;
   int n,m,nvt,nlp;
   vlist_t vls;
   vtype_t *vt;
   tupel_t *tp;
   tplist_t *cls;
   rcls_t *rc;
   ltree_t **lta;
   cnum_t **cna;
   tlist_t *tl;

   rc=trs[(int)(th)].rc;
   nvt=(*rc).nvt;
   vt=(*rc).vt;

   vls=vertices(th,order);
   cls=alloc_tplist(order);
   tp=NULL;

   for (i=1;i<=order;i++)
   {
      nlp=nlp_max(th,i);

      for (j=0;j<=nlp;j++)
      {
         tl=select_ltrees(th,i,j);

         if (tl!=NULL)
         {
            n=(*tl).n;
            lta=(*tl).lta;
            cna=(*tl).cna;

            for (k=0;k<n;k++)
            {
               if (is_zero_cnum(cna[k])==0)
               {
                  if (tp==NULL)
                     tp=alloc_tp(nvt+1);

                  (*tp).i[0]=i;

                  for (iv=0;iv<nvt;iv++)
                     (*tp).i[iv+1]=nVX(lta[k][0].dm,vt[iv],&vls);

                  m=(*cls).n;

                  if (add_tp(tp,cls)==m)
                     tp=NULL;
               }
            }
         }
      }
   }

   if (tp!=NULL)
      free_tp(tp);
   sort_tplist(cls);
   (*rc).cls=cls;
}


static void alloc_trs(theory_t th,int order)
{
   int ncl,i;
   rlist_t **rl;
   tsets_t *ts;

   ts=trs+(int)(th);
   ncl=(*ts).rc[0].cls[0].n;
   rl=malloc(ncl*sizeof(*rl));
   error_loc(rl==NULL,1,"alloc_trs [rtree_gen.c]",
             "Unable to allocate diagram lists");

   for (i=0;i<ncl;i++)
      rl[i]=alloc_rlist(32);

   (*ts).order=order;
   (*ts).rl=rl;
}


static void set_trs(theory_t th,int order)
{
   int i,j,k,iv;
   int n,nvt,nlp,icl;
   vlist_t vls;
   vtype_t *vt;
   tupel_t *tp;
   tplist_t *cls;
   rcls_t *rc;
   rtree_t *rt;
   rlist_t **rl;
   ltree_t **lta;
   cnum_t **cna;
   tlist_t *tl;
   tsets_t *ts;

   ts=trs+(int)(th);
   rc=(*ts).rc;
   rl=(*ts).rl;

   nvt=(*rc).nvt;
   vt=(*rc).vt;
   cls=(*rc).cls;

   vls=vertices(th,order);
   tp=alloc_tp(nvt+1);

   for (i=1;i<=order;i++)
   {
      nlp=nlp_max(th,i);

      for (j=0;j<=nlp;j++)
      {
         tl=select_ltrees(th,i,j);

         if (tl!=NULL)
         {
            n=(*tl).n;
            lta=(*tl).lta;
            cna=(*tl).cna;

            for (k=0;k<n;k++)
            {
               if (is_zero_cnum(cna[k])==0)
               {
                  (*tp).i[0]=i;

                  for (iv=0;iv<nvt;iv++)
                     (*tp).i[iv+1]=nVX(lta[k][0].dm,vt[iv],&vls);

                  icl=fnd_tp(tp,cls);
                  rt=get_rtree(th,order,lta[k]);
                  add_rtd(rt,cna[k],rl[icl]);
               }
            }
         }
      }
   }

   free_tp(tp);
}


static int nall_dgm(theory_t th)
{
   int nall,ncl,i;
   rlist_t **rl;
   tsets_t *ts;

   ts=trs+(int)(th);
   ncl=(*ts).rc[0].cls[0].n;
   rl=(*ts).rl;
   nall=0;

   for (i=0;i<ncl;i++)
      nall+=rl[i][0].n;

   return nall;
}


int rtrees(theory_t th,int order)
{
   tsets_t *ts;

   error_loc((th==THEORIES)||(order<=0),1,"rtrees [rtree_gen.c]",
             "Arguments are out of range");

   ts=trs+(int)(th);
   error_loc(((*ts).order>0)&&(order!=(*ts).order),1,"rtrees [rtree_gen.c]",
             "Attempt to modify the list of converted diagrams");

   if ((*ts).order==0)
   {
      set_vt(th,order);
      set_cls(th,order);
      alloc_trs(th,order);
      set_trs(th,order);
   }

   return nall_dgm(th);
}


rcls_t *rtree_classes(theory_t th)
{
   tsets_t *ts;

   error_loc(th==THEORIES,1,"rtree_classes [rtree_gen.c]",
             "Argument is out of range");

   ts=trs+(int)(th);
   error_loc((*ts).order==0,1,"rtree_classes [rtree_gen.c]",
             "rtrees() has not yet been called");

   return (*ts).rc;
}


rlist_t *select_rtrees(theory_t th,int icl)
{
   int ncl;
   tsets_t *ts;

   error_loc(th==THEORIES,1,"select_rtrees [rtree_gen.c]",
             "Argument 'th' is out of range");

   ts=trs+(int)(th);
   error_loc((*ts).order==0,1,"select_rtrees [rtree_gen.c]",
             "rtrees() has not yet been called");

   ncl=(*ts).rc[0].cls[0].n;
   error_loc((icl<0)||(icl>=ncl),1,"select_rtrees [rtree_gen.c]",
             "Argument 'icl' is out of range");

   return (*ts).rl[icl];
}
