
/*******************************************************************************
*
* File tplist.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Utility programs for lists of tupels.
*
*   tplist_t *alloc_tplist(int nmax)
*     Creates a new tupel list, with initially space for nmax tupels.
*     The program returns the pointer to the new list.
*
*   void sort_tplist(tplist_t *tpl)
*     Sorts the elements of the tupel list "tpl" in ascending order. The
*     ordering relation is the one described in tupel.c.
*
*   int add_tp(tupel_t *tp,tplist_t *tpl)
*     Appends the tupel "tp" to the tupel list "tpl" if it is not found
*     in the list. The program returns the index of the tupel in the list.
*
*   int fnd_tp(tupel_t *tp,tplist_t *tpl)
*     Returns the index of the tupel "tp" in the tupel list "tpl". If the
*     tupel not contained in the list, the program returns the number of
*     elements of the list.
*
* The structures of type tplist_t have the following elements:
*
*  n                     Current number of tupels included in the
*                        list.
*
*  nmax                  Current maximal number of tupels that can be
*                        included in the list. nmax and the arrays in
*                        the list are adjusted dynamically by add_tp().
*
*  tpa[i]                Pointer to the tupel number i (i=0,..,n-1).
*
* Any unused elements of the tupel array tpa are set to NULL. Note that the
* array address (*tpl).tpa may change when a tupel is added to the list.
*
* Currently no sophisticated algorithm is used for the searching and sorting
* of tupel lists. Tupel lists should therefore be kept fairly small.
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define TPLIST_C

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "utils.h"
#include "tupel.h"


static tupel_t **new_tpa(int nmax,int nold,tupel_t **tpold)
{
   int i;
   tupel_t **tpa;

   tpa=malloc(nmax*sizeof(*tpa));
   error_loc(tpa==NULL,1,"new_tpa [tplist.c]",
             "Unable to allocate tupel array");

   for (i=0;i<nold;i++)
      tpa[i]=tpold[i];

   for (i=nold;i<nmax;i++)
      tpa[i]=NULL;

   if (nold>0)
      free(tpold);

   return tpa;
}


tplist_t *alloc_tplist(int nmax)
{
   tplist_t *tpl;

   error_loc(nmax<1,1,"alloc_tplist [tplist.c]",
             "Argument nmax is out of range");

   tpl=malloc(1*sizeof(*tpl));
   error_loc(tpl==NULL,1,"alloc_tplist [tplist.c]",
             "Unable to allocate tupel list");

   (*tpl).n=0;
   (*tpl).nmax=nmax;
   (*tpl).tpa=new_tpa(nmax,0,NULL);

   return tpl;
}


void sort_tplist(tplist_t *tpl)
{
   int i,j,n;
   tupel_t **tpa,*tpx,*tpy;

   n=(*tpl).n;
   tpa=(*tpl).tpa;

   for (i=0;i<(n-1);i++)
   {
      tpx=tpa[i];

      for (j=(i+1);j<n;j++)
      {
         tpy=tpa[j];

         if (cmp_tp(tpy,tpx)==1)
         {
            tpa[i]=tpy;
            tpa[j]=tpx;
            tpx=tpy;
         }
      }
   }
}


int add_tp(tupel_t *tp,tplist_t *tpl)
{
   int m,n,nmax;
   tupel_t **tpa;

   n=(*tpl).n;
   tpa=(*tpl).tpa;

   for (m=0;m<n;m++)
   {
      if (cmp_tp(tp,tpa[m])==0)
         return m;
   }

   nmax=(*tpl).nmax;

   if (n==nmax)
   {
      if (nmax<=(INT_MAX/2))
         m=2*nmax;
      else
      {
         error_loc(nmax==INT_MAX,1,"add_tp [tplist.c]",
                   "Unable to extend tupel list");
         m=INT_MAX;
      }

      (*tpl).nmax=m;
      (*tpl).tpa=new_tpa(m,nmax,(*tpl).tpa);
   }

   (*tpl).n=n+1;
   (*tpl).tpa[n]=tp;

   return n;
}


int fnd_tp(tupel_t *tp,tplist_t *tpl)
{
   int m,n;
   tupel_t **tpa;

   n=(*tpl).n;
   tpa=(*tpl).tpa;

   for (m=0;m<n;m++)
   {
      if (cmp_tp(tp,tpa[m])==0)
         return m;
   }

   return n;
}
