
/*******************************************************************************
*
* File tupel.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Utility programs for tupels.
*
*   tupel_t *alloc_tp(int n)
*     Returns the pointer to a newly allocated tupel with "n" elements.
*     All elements are set to 0. An error occurs if the allocation fails.
*
*   void free_tp(tupel_t *tp)
*     Frees the integer array in the structure "tp" and then the structure
*     itself. The program assumes that "tp" was allocated by alloc_tp().
*
*   void int cmp_tp(tupel_t *tp1,tupel_t *tp2)
*     Returns 1 if tp1<tp2, 0 if tp1=tp2 and -1 if tp1>tp2, where the
*     ordering of the tupels is lexicographic (see the notes).
*
* A tupel is an ordered list of n>=0 integers. The empty tupel tp0 has
* tp0.n=0 and tp0.i=NULL. Two tupels tp1 and tp2 satisfy tp1<tp2 if and
* only if
*
*   tp1.n<tp2.n
*
* or
*
*   tp1.n=tp2.n
*
* and there exists some k in [0,tp1.n) such that
*
*   tp1.i[k]<tp2.i[k] and tp1.i[j]=tp2.i[j] for all j<k.
*
* This defines a total ordering in the set of all tupels.
*
* The programs in this module do not perform any communications and can be
* locally called.
*
*******************************************************************************/

#define TUPEL_C

#include <stdlib.h>
#include <stdio.h>
#include "utils.h"
#include "tupel.h"


tupel_t *alloc_tp(int n)
{
   int *i,*im;
   tupel_t *tp;

   error_loc(n<0,1,"alloc_tp [tupel.c]","Argument is out of range");

   tp=malloc(1*sizeof(*tp));

   if (n>0)
      i=malloc(n*sizeof(*i));
   else
      i=NULL;

   error_loc((tp==NULL)||((n>0)&&(i==NULL)),1,
             "alloc_tp [tupel.c]","Unable to allocate structure");

   (*tp).n=n;
   (*tp).i=i;
   im=i+n;

   for (;i<im;i++)
      (*i)=0;

   return tp;
}


void free_tp(tupel_t *tp)
{
   if (tp!=NULL)
   {
      if ((*tp).n>0)
         free((*tp).i);

      free(tp);
   }
}


int cmp_tp(tupel_t *tp1,tupel_t *tp2)
{
   int n1,n2,*i1,*i2,*i1m;

   n1=(*tp1).n;
   n2=(*tp2).n;

   if (n1<n2)
      return 1;
   else if (n1>n2)
      return -1;

   i1=(*tp1).i;
   i2=(*tp2).i;
   i1m=i1+n1;

   for (;i1<i1m;i1++)
   {
      if ((*i1)<(*i2))
         return 1;
      else if ((*i1)>(*i2))
         return -1;

      i2+=1;
   }

   return 0;
}
