
/*******************************************************************************
*
* File AFPn.c
*
* Copyright (C) 2015, 2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of the vertex functions A_FP^n.
*
*   void AFPn(int n,sl3_alg_dble **asrc,sl3_alg_dble *cbsrc,
*             sl3_alg_dble csrc,sl3_alg_dble *vxfld)
*     Assigns the vertex function AFP^n to the field vxfld for given
*     source fields asrc[0],..,asrc[n-2], cbsrc and csrc (representing
*     the gauge, antighost and ghost fields, respectively). All source
*     fields are assumed to satisfy the boundary conditions specified
*     in the parameter data base.
*
* See the notes
*
*  M. Luescher: "Lattice QCD vertices in ISPT", file doc/vxfcts.pdf,
*
* for the definition of the vertex functions A_FP^n and a description of
* the computational strategy employed.
*
* Gauge potentials are locally stored as described in main/README.global
* and aflds/README.aflds. Both input and output fields must be of this
* type. The output field may not overlap with any of the input fields.
* The ghost fields are locally stored as described in main/README.global
* and cflds/README.cflds.
*
* The parameter n must be either equal to 1 or an even number in the range
* 2<=n<=NMAX, where NMAX is defined in vxfcts.h.
*
* The program in this module acts globally and must be called on all MPI
* processes simultaneously.
*
*******************************************************************************/

#define AFPN_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "sl3.h"
#include "utils.h"
#include "lattice.h"
#include "su3fcts.h"
#include "linalg.h"
#include "aflds.h"
#include "cflds.h"
#include "vxfcts.h"
#include "global.h"

#define N0 (NPROC0*L0)

static int ns=0;
static double ifact[NMAX+1];
static const su3_dble ud0={{0.0,0.0},{0.0,0.0},{0.0,0.0},
                           {0.0,0.0},{0.0,0.0},{0.0,0.0},
                           {0.0,0.0},{0.0,0.0},{0.0,0.0}};
static su3_dble As ALIGNED16;
static su3_dble Cb ALIGNED16;
static su3_dble C ALIGNED16;
static su3_dble dCb ALIGNED16;
static su3_dble dC ALIGNED16;
static su3_dble Xw ALIGNED16;
static su3_dble *Ak,*dPV,*CbdPV,*CdPV,*dPZ;


static void alloc_arrays(int n)
{
   int nd,i;

   nd=1<<(n-1);

   if (n>ns)
   {
      if (ns==0)
      {
         ifact[0]=1.0;

         for (i=1;i<=NMAX;i++)
            ifact[i]=ifact[i-1]/(double)(i);
      }

      if (ns>0)
         afree(Ak);

      Ak=amalloc((n-1+4*nd)*sizeof(*Ak),ALIGN);
      error(Ak==NULL,1,"alloc_arrays [AFPn.c]",
            "Unable to allocate auxiliary arrays");

      cm3x3_zero(n-1+4*nd,Ak);
      ns=n;
   }

   dPV=Ak+n-1;
   CbdPV=dPV+nd;
   CdPV=CbdPV+nd;
   dPZ=CdPV+nd;
}


static void set_dPV(int n,int ip,sl3_alg_dble **asrc)
{
   int i;

   for (i=0;i<(n-1);i++)
      sl3_to_su3(asrc[i]+ip,Ak+i);

   sym_prod(2,n-1,Ak,dPV);
}


static void set_CbdPV(int n)
{
   int ip,ipmx;

   ipmx=1<<(n-1);

   for (ip=0;ip<ipmx;ip++)
   {
      su3xsu3(&dCb,dPV+ip,CbdPV+ip);
      su3xsu3(&dC,dPV+ip,CdPV+ip);
   }
}


static void set_dPZ(int n)
{
   int ip,iq,ipmx,iqmx;
   int p,q,iuq,iur;
   double r;

   ipmx=1<<(n-1);

   for (ip=0;ip<ipmx;ip++)
   {
      dPZ[ip]=ud0;
      p=bcount(n-1,ip);
      iqmx=1<<p;

      for (iq=0;iq<iqmx;iq++)
      {
         q=bcount(p,iq);
         r=ifact[q]*ifact[n-q];
         if (q&0x1)
            r=-r;
         iuq=bsubset(n-1,iq,ip);
         iur=ip^iuq;

         su3xsu3(CdPV+iuq,CbdPV+iur,&Xw);
         cm3x3_mul_add(CbdPV+iuq,CdPV+iur,&Xw);
         cm3x3_mulr_add(&r,&Xw,dPZ+ip);
      }
   }
}


static void last_prod(int n)
{
   int ir,is,irmx;

   irmx=1<<(n-1);
   As=dPZ[irmx-1];

   for (ir=1;ir<irmx;ir++)
   {
      is=(irmx-1)^ir;
      cm3x3_mul_add(dPV+ir,dPZ+is,&As);
   }
}


void AFPn(int n,sl3_alg_dble **asrc,sl3_alg_dble *cbsrc,
          sl3_alg_dble *csrc,sl3_alg_dble *vxfld)
{
   int iprms[1];
   int ix,iy,t,ifc,mu,ip;
   double wc,half,bn;

   error_root((n<1)||(n>NMAX)||((n>1)&&((n%2)==1)),1,"AFPn [AFPn.c]",
              "Parameter n is out of range");

   if (NPROC>1)
   {
      iprms[0]=n;
      MPI_Bcast(iprms,1,MPI_INT,0,MPI_COMM_WORLD);
      error(iprms[0]!=n,1,"AFPn [AFPn.c]","Parameter n is not global");
   }

   alloc_arrays(n);
   cpcd_int_bnd(0,cbsrc);
   cpcd_int_bnd(0,csrc);
   half=0.5;
   bn=bernoulli(n);
   ip=0;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         mu=ifc/2;

         if (ifc&0x1)
         {
            if ((t>0)||(mu>0))
            {
               sl3_to_su3(cbsrc+ix,&dCb);
               sl3_to_su3(csrc+ix,&dC);
            }
            else
            {
               dCb=ud0;
               dC=ud0;
            }

            iy=idn[ix][mu];
            sl3_to_su3(cbsrc+iy,&Cb);
            sl3_to_su3(csrc+iy,&C);
            cm3x3_sub(&Cb,&dCb);
            cm3x3_sub(&C,&dC);
         }
         else
         {
            if ((t<(N0-1))||(mu>0))
            {
               iy=iup[ix][mu];
               sl3_to_su3(cbsrc+iy,&dCb);
               sl3_to_su3(csrc+iy,&dC);
            }
            else
            {
               dCb=ud0;
               dC=ud0;
            }

            sl3_to_su3(cbsrc+ix,&Cb);
            sl3_to_su3(csrc+ix,&C);
            cm3x3_sub(&Cb,&dCb);
            cm3x3_sub(&C,&dC);
         }

         if (n==1)
         {
            cm3x3_mulr(&half,&dC,&dC);
            cm3x3_add(&C,&dC);
            su3xsu3(&dCb,&dC,&As);
            cm3x3_mul_sub(&dC,&dCb,&As);
            wc=1.0;
         }
         else
         {
            set_dPV(n,ip,asrc);
            set_CbdPV(n);
            set_dPZ(n);
            last_prod(n);
            wc=-bn;
         }

         if ((t==0)&&(mu>0))
            wc*=0.5;

         cm3x3_mulr(&wc,&As,&As);
         su3_to_sl3(&As,vxfld+ip);
         ip+=1;
      }
   }

   if (cpr[0]==(NPROC0-1))
      set_calg2zero(3,vxfld+4*VOLUME+7*(BNDRY/4));
}
