
/*******************************************************************************
*
* File vxgen.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Utility programs for the computation of lattice QCD vertex functions.
*
*   int bcount(int n,unsigned int x)
*     Examines the lowest n bits of x and returns the number of 1's
*     found.
*
*   unsigned int binsert(int n,unsigned int x,unsigned int y)
*     Examines the lowest n bits of y and returns y with the 0's found
*     replaced by the lowest bits of x (see the notes).
*
*   unsigned int bsubset(int n,unsigned int x,unsigned int y)
*     Examines the lowest n bits of y and returns y with the 1's found
*     replaced by the lowest bits of x (see the notes).
*
*   double bernoulli(int n)
*     Returns the Bernoulli number B_n.
*
*   void sym_prod(int type,int n,su3_dble *X,su3_dble *PX)
*     Computes the symmetric products PX[0],..,PX[2^n-1] of the matrices
*     X[0],..,X[n-1]. PX[0] is set to the unit matrix. If type=1,2 or 3,
*     the products are multiplied by (-1)^p, p! or (-1)^p*p!, where p is
*     the number of factors in the product (see the notes).
*
* The programs bcount(), binsert() and bsubset() take it for granted that
*
*  n<=8*sizeof(unsigned int).
*
* The output of these programs may be incorrect if this condition is not met.
*
* In the case of the programs bernoulli() and sym_prod(), n must be in the
* range 0<=n<=NMAX, where NMAX is defined in vxfcts.h. An error occurs if
* n is not in this range.
*
* The program binsert() goes through the bits of y, starting with the lowest
* one, and replaces the 0's found one by one by the bits of x. If there are
* k zeros, the first k bits of x are inserted into y. The program bsubset()
* does the same except that the 1's in y are replaced instead of the 0's.
*
* For the Bernoulli numbers, the Seidel formula quoted in
*
*  http://ami.ektf.hu/uploads/papers/finalpdf/AMI_38_from123to126.pdf
*
* is used. This method yields the Bernoulli numbers practically to machine
* precision.
*
* The symmetric products computed by sym_prod() are indexed by an integer
* idx running from 0 to 2^n-1. If
*
*  idx=b[0]*2^0+b[1]*2^1+..+b[n-1]*2^(n-1),   b[k]=0 or 1,
*
* is the binary representation of the index, the product PX[idx] is the one
* of the factors X[k] where b[k]=1. In particular, the number of factors in
* the product is equal to the bit count of the index idx.
*
* See appendix B of the notes
*
*  M. Luescher: "Lattice QCD vertices in ISPT", file doc/vxfcts.pdf,
*
* for the definition and properties of the symmetric products.
*
* The programs in this module do not perform any global communications and
* may be locally called.
*
*******************************************************************************/

#define VXGEN_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "sl3.h"
#include "utils.h"
#include "su3fcts.h"
#include "vxfcts.h"

static int ns,init=0;
static int flag[1<<NMAX];
static unsigned int ids;
static double fact[NMAX+1],Bn[NMAX+1];
static su3_dble *PXs;


static void set_fact(void)
{
   int n,nh,k;
   double b;

   fact[0]=1.0;
   fact[1]=1.0;

   Bn[0]=1.0;
   Bn[1]=-0.5;

   for (n=2;n<=NMAX;n++)
   {
      fact[n]=1.0/(double)(n);
      Bn[n]=0.0;

      if ((n%2)==0)
      {
         nh=n/2;
         b=1.0/(double)(n+1);

         for (k=0;k<nh;k++)
         {
            Bn[n]-=b*Bn[nh+k];
            b*=(double)((nh+1-k)*(nh+k+2))/(double)((k+1)*(nh+k+1));
         }
      }
   }

   init=1;
}


int bcount(int n,unsigned int x)
{
   int m;

   m=0;

   for (;n>0;n--)
   {
      m+=(x&1u);
      x>>=1u;
   }

   return m;
}


unsigned int binsert(int n,unsigned int x,unsigned int y)
{
   unsigned int p;

   p=1u;

   for (;n>0;n--)
   {
      if (p&y)
         x<<=1u;
      else
         y|=(p&x);

      p<<=1u;
   }

   return y;
}


unsigned int bsubset(int n,unsigned int x,unsigned int y)
{
   unsigned int p;

   p=1u;
   x=~x;

   for (;n>0;n--)
   {
      if (p&y)
         y^=(p&x);
      else
         x<<=1u;

      p<<=1u;
   }

   return y;
}


double bernoulli(int n)
{
   if (init==0)
      set_fact();

   if ((n>=0)&&(n<=NMAX))
      return Bn[n];
   else
   {
      error_loc(1,1,"bernoulli [vxgen.c]",
                "Parameter n is out of range");
      return 0.0;
   }
}


static su3_dble *rsp0(unsigned int idx,int p)
{
   int is;
   unsigned int id1,idy;
   su3_dble *PX;

   PX=PXs+idx;
   flag[idx]=1;
   is=0;

   for (id1=1u;id1<ids;id1<<=1u)
   {
      if (idx&id1)
      {
         idy=idx^id1;

         if (flag[idy])
         {
            if (is)
               cm3x3_mul_add(PXs+id1,PXs+idy,PX);
            else
            {
               su3xsu3(PXs+id1,PXs+idy,PX);
               is=1;
            }
         }
         else
         {
            if (is)
               cm3x3_mul_add(PXs+id1,rsp0(idy,p-1),PX);
            else
            {
               su3xsu3(PXs+id1,rsp0(idy,p-1),PX);
               is=1;
            }
         }
      }
   }

   cm3x3_mulr(fact+p,PX,PX);

   return PX;
}


static void sp0(su3_dble *X)
{
   unsigned int id1;

   for (id1=1u;id1<ids;id1<<=1u)
   {
      cm3x3_assign(1,X,PXs+id1);
      flag[id1]=1;
      X+=1;
   }

   if (ns>=2)
      (void)(rsp0(ids-1u,ns));
}


static void sp1(su3_dble *X)
{
   unsigned int id1;
   double r;

   r=-1.0;

   for (id1=1u;id1<ids;id1<<=1u)
   {
      cm3x3_mulr(&r,X,PXs+id1);
      flag[id1]=1;
      X+=1;
   }

   if (ns>=2)
      (void)(rsp0(ids-1u,ns));
}


static su3_dble *rsp2(unsigned int idx)
{
   int is;
   unsigned int id1,idy;
   su3_dble *PX;

   PX=PXs+idx;
   flag[idx]=1;
   is=0;

   for (id1=1u;id1<ids;id1<<=1u)
   {
      if (idx&id1)
      {
         idy=idx^id1;

         if (flag[idy])
         {
            if (is)
               cm3x3_mul_add(PXs+id1,PXs+idy,PX);
            else
            {
               su3xsu3(PXs+id1,PXs+idy,PX);
               is=1;
            }
         }
         else
         {
            if (is)
               cm3x3_mul_add(PXs+id1,rsp2(idy),PX);
            else
            {
               su3xsu3(PXs+id1,rsp2(idy),PX);
               is=1;
            }
         }
      }
   }

   return PX;
}


static void sp2(su3_dble *X)
{
   unsigned int id1;

   for (id1=1u;id1<ids;id1<<=1u)
   {
      cm3x3_assign(1,X,PXs+id1);
      flag[id1]=1;
      X+=1;
   }

   if (ns>=2)
      (void)(rsp2(ids-1u));
}


static void sp3(su3_dble *X)
{
   unsigned int id1;
   double r;

   r=-1.0;

   for (id1=1u;id1<ids;id1<<=1u)
   {
      cm3x3_mulr(&r,X,PXs+id1);
      flag[id1]=1;
      X+=1;
   }

   if (ns>=2)
      (void)(rsp2(ids-1u));
}


void sym_prod(int type,int n,su3_dble *X,su3_dble *PX)
{
   int i;

   if ((n>=0)&&(n<=NMAX))
   {
      cm3x3_unity(1,PX);

      if (n>=1)
      {
         if (init==0)
            set_fact();

         ns=n;
         ids=1u<<n;
         PXs=PX;
         flag[0]=1;

         for (i=1;i<(int)(ids);i++)
            flag[i]=0;

         if (type==1)
            sp1(X);
         else if (type==2)
            sp2(X);
         else if (type==3)
            sp3(X);
         else
            sp0(X);
      }
   }
   else
      error_loc(1,1,"sym_prod [vxgen.c]","Parameter n is out of range");
}
