
/*******************************************************************************
*
* File wforce.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of the force driving the gradient flow.
*
*   void plaq_frc(int n,su3_dble *ud,sl3_alg_dble *fd)
*     Assigns the force deriving from the Wilson plaquette action to the
*     force field fd (see the notes).
*
*   void damping_frc(int n,su3_dble *ud,sl3_alg_dble *fd)
*     *Adds* the gauge-damping force to the force field fd (see the notes).
*
* The implementation of the gradient flow is described in
*
*   M. Luescher, "Yang-Mills gradient flow in stochastic perturbation
*                 theory", notes (May 2015), file doc/gflow.pdf.
*
* In all programs in this module, the array ud is expected to contain the
* n+1 coefficient fields in the expansion
*
*   U=1+g0*U[0]+g0^2*U[1]+..+g0^(n+1)*U[n]
*
* of the gauge field in powers of the coupling g0 (cf. uflds/uflds.c). The
* array fd similarly contains the n+1 coefficient fields in the expansion
*
*   F=g0*F[0]+g0^2*F[1]+..+g0^(n+1)*F[n]
*
* of the force field.
*
* On the links in the local lattice, where the static link variables reside,
* the programs set the force field to zero.
*
* The program damping_frc() requires a workspace of n+1 sl(3,C) link fields
* and n+1 sl(3,C) ghost fields.
*
* The programs in this module perform global communications and must be
* called simultaneously on all MPI processes.
*
*******************************************************************************/

#define WFORCE_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "pmdb.h"
#include "su3fcts.h"
#include "utils.h"
#include "lattice.h"
#include "aflds.h"
#include "cflds.h"
#include "uflds.h"
#include "linalg.h"
#include "wflow.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)
#define NMAX 16

static int ms=0;
static int nsm[NMAX],ism[NMAX];
static double *rt;
static sl3_alg_dble *Ct0,*Aw;
static su3_dble *Vw,*Ww,*Xw,*Yw,*Uw[4];


static void alloc_arrays(int n)
{
   int m,i;
   double r;

   if (ms==0)
   {
      rt=malloc(N0*sizeof(*rt));
      error(rt==NULL,1,"alloc_arrays [wforce.c]",
            "Unable to allocate auxiliary arrays");

      for (i=0;i<NMAX;i++)
      {
         nsm[i]=0;
         ism[i]=0;
      }

      r=1.0/((double)(N0*N0)*(double)(N1*N2*N3));

      for (i=0;i<N0;i++)
         rt[i]=(1.0-(double)(i)/(double)(N0))*r;
   }

   if (ms>0)
   {
      afree(Ct0);
      afree(Vw);
   }

   m=n+1;
   Ct0=amalloc(2*m*sizeof(*Ct0),ALIGN);
   Vw=amalloc(8*m*sizeof(*Vw),ALIGN);
   error((Ct0==NULL)||(Vw==NULL),2,"alloc_arrays [wforce.c]",
         "Unable to allocate auxiliary arrays");
   Aw=Ct0+m;
   Ww=Vw+m;
   Xw=Ww+m;
   Yw=Xw+m;
   Uw[0]=Yw+m;

   for (i=1;i<4;i++)
      Uw[i]=Uw[i-1]+m;

   set_calg2zero(2*m,Ct0);
   cm3x3_zero(8*m,Vw);
   ms=m;
}


static int set_ism(int n)
{
   int m,i;

   m=16*(n+1);

   for (i=0;i<NMAX;i++)
   {
      if (nsm[i]==0)
      {
         nsm[i]=m;
         ism[i]=init_hsum(m);
      }

      if (nsm[i]==m)
         return ism[i];
   }

   error(1,1,"set_ism [wforce.c]","Requested too many hsum instances");

   return 0;
}


void plaq_frc(int n,su3_dble *ud,sl3_alg_dble *fd)
{
   int bc,m,i;
   int ix,t,pln,ip[4];
   double r0,r1;

   bc=bc_type();
   m=n+1;

   if (m>ms)
      alloc_arrays(n);
   copy_bnd_ud(n,ud);

   if (cpr[0]==(NPROC0-1))
   {
      cm3x3_zero(3*m,ud+(4*VOLUME+7*(BNDRY/4))*m);
      set_calg2zero((4*VOLUME+7*(BNDRY/4)+3)*m,fd);
   }
   else
      set_calg2zero((4*VOLUME+7*(BNDRY/4))*m,fd);

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (pln=0;pln<6;pln++)
      {
         if ((t!=0)||(pln<3)||(bc!=1))
         {
            plaq_aidx(pln,ix,ip);

            cm3x3_assign(m,ud+m*ip[0],Uw[0]);
            cm3x3_assign(m,ud+m*ip[1],Uw[1]);
            cm3x3_inv_poly(m,ud+m*ip[2],Uw[2]);
            cm3x3_inv_poly(m,ud+m*ip[3],Uw[3]);

            cm3x3_varmul_poly(m,Uw[3],Uw[2],Vw);
            cm3x3_varmul_poly(m,Uw[1],Vw,Ww);
            cm3x3_varmul_poly(m,Uw[0],Ww,Vw);
            cm3x3_inv_poly(m,Vw,Ww);

            r0=-0.5;
            r1=0.5;

            if ((t==0)&&(pln<3))
            {
               if (bc==1)
                  r1=0.0;
               else
                  r1=1.0;
            }

            for (i=0;i<m;i++)
            {
               cm3x3_sub(Ww+i,Vw+i);
               su3_mulr_add_sl3(&r0,Vw+i,fd+m*ip[0]+i);

               if (r1!=0.0)
                  su3_mulr_add_sl3(&r1,Vw+i,fd+m*ip[2]+i);
            }

            cm3x3_varmul_poly(m,Uw[2],Uw[0],Vw);
            cm3x3_varmul_poly(m,Uw[1],Uw[3],Ww);
            cm3x3_varmul_poly(m,Ww,Vw,Xw);
            cm3x3_varmul_poly(m,Vw,Ww,Yw);
            cm3x3_inv_poly(m,Xw,Vw);
            cm3x3_inv_poly(m,Yw,Ww);

            r0=-0.5;
            r1=0.5;

            if ((t==(N0-1))&&(pln<3))
               r0=0.0;

            for (i=0;i<m;i++)
            {
               if (r0!=0.0)
               {
                  cm3x3_sub(Vw+i,Xw+i);
                  su3_mulr_add_sl3(&r0,Xw+i,fd+m*ip[1]+i);
               }

               cm3x3_sub(Ww+i,Yw+i);
               su3_mulr_add_sl3(&r1,Yw+i,fd+m*ip[3]+i);
            }
         }
      }
   }

   add_bnd_ad(n,fd);
}


static void set_Ct(int n,su3_dble *ud,sl3_alg_dble *ad)
{
   int bc,m,i;
   int is,ix,ifc;
   double r;

   bc=bc_type();
   m=n+1;
   r=-0.5;

   if (bc==1)
   {
      is=set_ism(n);
      reset_hsum(is);

      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         for (ifc=0;ifc<8;ifc++)
         {
            cm3x3_inv_poly(m,ud,Vw);

            for (i=0;i<m;i++)
            {
               cm3x3_sub(ud+i,Vw+i);
               cm3x3_mulr(&r,Vw+i,Vw+i);
               su3_to_sl3(Vw+i,ad+i);
            }

            if (ifc<2)
               add_to_hsum(is,(double*)(ad));

            ud+=m;
            ad+=m;
         }
      }

      global_hsum(is,(double*)(Ct0));
   }
   else
   {
      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         for (ifc=0;ifc<8;ifc++)
         {
            cm3x3_inv_poly(m,ud,Vw);

            for (i=0;i<m;i++)
            {
               cm3x3_sub(ud+i,Vw+i);
               cm3x3_mulr(&r,Vw+i,Vw+i);
               su3_to_sl3(Vw+i,ad+i);
            }

            ud+=m;
            ad+=m;
         }
      }
   }
}


static void set_omega(int n,sl3_alg_dble *ad,sl3_alg_dble *cd)
{
   int bc,m;
   int ix,iy,t,mu;

   bc=bc_type();
   m=n+1;

   if (bc==1)
   {
      for (ix=0;ix<(VOLUME/2);ix++)
      {
         t=global_time(ix);
         mulr_calg(m,rt[t],Ct0,cd+ix*m);
      }

      set_calg2zero((BNDRY/2)*m,cd+VOLUME*m);

      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         t=global_time(ix);

         mulr_calg(m,rt[t],Ct0,cd+ix*m);

         if (t!=0)
            add_calg(m,ad,cd+ix*m);

         if (t!=(N0-1))
         {
            iy=iup[ix][0];
            sub_calg(m,ad,cd+iy*m);
         }

         ad+=m;

         if (t!=0)
            sub_calg(m,ad,cd+ix*m);

         if (t!=1)
         {
            iy=idn[ix][0];
            add_calg(m,ad,cd+iy*m);
         }

         ad+=m;

         if (t!=0)
         {
            for (mu=1;mu<4;mu++)
            {
               add_calg(m,ad,cd+ix*m);
               iy=iup[ix][mu];
               sub_calg(m,ad,cd+iy*m);
               ad+=m;

               sub_calg(m,ad,cd+ix*m);
               iy=idn[ix][mu];
               add_calg(m,ad,cd+iy*m);
               ad+=m;
            }
         }
         else
            ad+=6*m;
      }
   }
   else
   {
      set_calg2zero((VOLUME/2)*m,cd);
      set_calg2zero((BNDRY/2)*m,cd+VOLUME*m);

      for (ix=(VOLUME/2);ix<VOLUME;ix++)
      {
         t=global_time(ix);

         assign_calg2calg(m,ad,cd+ix*m);

         if (t==0)
            add_calg(m,ad,cd+ix*m);

         if (t!=(N0-1))
         {
            iy=iup[ix][0];
            sub_calg(m,ad,cd+iy*m);
         }

         ad+=m;

         if (t!=0)
            sub_calg(m,ad,cd+ix*m);

         iy=idn[ix][0];
         add_calg(m,ad,cd+iy*m);

         if (t==1)
            add_calg(m,ad,cd+iy*m);

         ad+=m;

         for (mu=1;mu<4;mu++)
         {
            add_calg(m,ad,cd+ix*m);
            iy=iup[ix][mu];
            sub_calg(m,ad,cd+iy*m);
            ad+=m;

            sub_calg(m,ad,cd+ix*m);
            iy=idn[ix][mu];
            add_calg(m,ad,cd+iy*m);
            ad+=m;
         }
      }
   }

   cpcd_ext_bnd(n,cd);
}


static void convert_cd2ud(int n,sl3_alg_dble *cd,su3_dble *ud)
{
   int i;

   for (i=0;i<=n;i++)
   {
      sl3_to_su3(cd,ud);
      cd+=1;
      ud+=1;
   }
}


static void convert_ud2cd(int n,su3_dble *ud,sl3_alg_dble *cd)
{
   int i;

   for (i=0;i<=n;i++)
   {
      su3_to_sl3(ud,cd);
      ud+=1;
      cd+=1;
   }
}


void damping_frc(int n,su3_dble *ud,sl3_alg_dble *fd)
{
   int bc,m;
   int ix,iy,t,mu;
   sl3_alg_dble **wad,**wcd;

   bc=bc_type();
   m=n+1;

   if (m>ms)
      alloc_arrays(n);

   wad=reserve_wad(m);
   wcd=reserve_wcd(m);
   set_Ct(n,ud,wad[0]);
   set_omega(n,wad[0],wcd[0]);
   cpcd_int_bnd(n,wcd[0]);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      if (t!=(N0-1))
      {
         iy=iup[ix][0];
         convert_cd2ud(n,wcd[0]+iy*m,Vw);
         cm3x3_varmul_poly(m,ud,Vw,Ww);
         cm3x3_inv_poly(m,ud,Xw);
         cm3x3_varmul_poly(m,Ww,Xw,Yw);
         convert_ud2cd(n,Yw,Aw);
         add_calg(m,Aw,fd);
      }

      sub_calg(m,wcd[0]+ix*m,fd);
      ud+=m;
      fd+=m;

      if (t!=0)
      {
         convert_cd2ud(n,wcd[0]+ix*m,Vw);
         cm3x3_varmul_poly(m,ud,Vw,Ww);
         cm3x3_inv_poly(m,ud,Xw);
         cm3x3_varmul_poly(m,Ww,Xw,Yw);
         convert_ud2cd(n,Yw,Aw);
         add_calg(m,Aw,fd);
      }

      iy=idn[ix][0];
      sub_calg(m,wcd[0]+iy*m,fd);
      ud+=m;
      fd+=m;

      if ((bc!=1)||(t!=0))
      {
         for (mu=1;mu<4;mu++)
         {
            iy=iup[ix][mu];
            convert_cd2ud(n,wcd[0]+iy*m,Vw);
            cm3x3_varmul_poly(m,ud,Vw,Ww);
            cm3x3_inv_poly(m,ud,Xw);
            cm3x3_varmul_poly(m,Ww,Xw,Yw);
            convert_ud2cd(n,Yw,Aw);
            add_calg(m,Aw,fd);

            sub_calg(m,wcd[0]+ix*m,fd);
            ud+=m;
            fd+=m;

            convert_cd2ud(n,wcd[0]+ix*m,Vw);
            cm3x3_varmul_poly(m,ud,Vw,Ww);
            cm3x3_inv_poly(m,ud,Xw);
            cm3x3_varmul_poly(m,Ww,Xw,Yw);
            convert_ud2cd(n,Yw,Aw);
            add_calg(m,Aw,fd);

            iy=idn[ix][mu];
            sub_calg(m,wcd[0]+iy*m,fd);
            ud+=m;
            fd+=m;
         }
      }
      else
      {
         ud+=6*m;
         fd+=6*m;
      }
   }

   release_wcd();
   release_wad();
}
